#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from fabric.actor.core.common.exceptions import ResourcesException
from fabric.actor.core.common.resource_vector import ResourceVector
from fabric.actor.core.apis.i_resource_ticket_factory import IResourceTicketFactory
from fabric.actor.core.delegation.resource_delegation import ResourceDelegation
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_type import ResourceType


class ResourceTicket:
    """
    ResourceTicket represents a sequence of one or more delegations of resources from one party to another.
    """

    def __init__(self, *, factory: IResourceTicketFactory, delegation: ResourceDelegation, source):
        """
        Creates a root ticket from the specified delegation.
        @params delegation: root delegation
        @params factory: factory
        """
        if factory is None:
            raise ResourcesException("factory Null")
        if delegation is None:
            raise ResourcesException("delegation Null")

        self.factory = factory
        # All delegation records represented by this ticket.
        self.delegations = []
        if delegation is not None:
            self.delegations.append(delegation)

        if source is not None:
            if not isinstance(source, ResourceTicket):
                raise ResourcesException("Invalid argument")

            for d in source.delegations:
                self.delegations.append(d)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['factory']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.factory = None

    def __str__(self):
        result = "factory: {}".format(self.factory)
        for d in self.delegations:
            result += " {}".format(d)
        return result

    def is_valid(self) -> bool:
        """
        Is ticket valid
        @return true if valid, false otherwise
        """
        if self.delegations is None or len(self.delegations) == 0:
            return False

        for i in range(len(self.delegations)):
            if not self.delegations[i].is_valid():
                return False

        return True

    def get_delegation(self) -> ResourceDelegation:
        """
        Get Resource Delegation
        @return delegation
        """
        return self.delegations[0]

    def get_resource_type(self) -> ResourceType:
        """
        Get Resource Type
        @return delegation resource type
        """
        return self.get_delegation().get_resource_type()

    def get_properties(self) -> dict:
        """
        Get Delegation Properties
        @return delegation properties
        """
        return self.get_delegation().get_properties()

    def get_issuer(self) -> ID:
        """
        Get Issuer Guid
        @return issuer guid
        """
        return self.get_delegation().get_issuer()

    def get_holder(self) -> ID:
        """
        Get Delegation Holder
        @return delegation holder
        """
        return self.get_delegation().get_holder()

    def get_guid(self) -> ID:
        """
        Get Delegation Id
        @return delegation id
        """
        return self.get_delegation().get_guid()

    def get_term(self) -> Term:
        """
        Get Delegation Term
        @return delegation term
        """
        return self.get_delegation().get_term()

    def get_units(self) -> int:
        """
        Get number of units
        @return number of units
        """
        return self.get_delegation().get_units()

    def get_resource_vector(self) -> ResourceVector:
        """
        Get Resource vector
        @return resource vector
        """
        return self.get_delegation().get_resource_vector()

    def get_factory(self) -> IResourceTicketFactory:
        """
        Get Ticket factory
        @return ticket factory
        """
        return self.factory

    def set_factory(self, *, factory: IResourceTicketFactory):
        """
        Set ticket factory
        @param factory ticket factory
        """
        self.factory = factory



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import unittest
from datetime import datetime
import time
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.delegation.resource_bin import ResourceBin
from fabric.actor.core.delegation.simple_resource_ticket_factory import SimpleResourceTicketFactory
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.test.base_test_case import BaseTestCase


class SimpleTicketTest(BaseTestCase, unittest.TestCase):
    from fabric.actor.core.container.globals import Globals
    Globals.config_file = Constants.test_vm_am_configuration_file

    from fabric.actor.core.container.globals import GlobalsSingleton
    GlobalsSingleton.get().start(force_fresh=True)
    while not GlobalsSingleton.get().start_completed:
        time.sleep(0.0001)

    def setUp(self) -> None:
        self.factory = self.make_ticket_factory()
        self.now = datetime.utcnow()
        self.end = self.now.replace(day=1)
        self.term = Term(start=self.now, end=self.end)
        self.type = ResourceType(resource_type=str(ID()))
        self.units = 100
        self.create_factories()

    def make_ticket_factory(self) -> SimpleResourceTicketFactory:
        return SimpleResourceTicketFactory()

    def create_factories(self):
        actor = self.get_actor(name="site0", guid=ID())
        self.factory = self.make_ticket_factory()
        self.factory.set_actor(actor=actor)
        self.factory.initialize()

    def test_simple_ticket(self):
        d0 = self.factory.make_delegation(units=self.units, term=self.term, rtype=self.type)
        ticket0 = self.factory.make_ticket(delegation=d0)

        print(self.factory.to_json(ticket=ticket0))

    def test_simple_ticket_with_bins(self):
        bin0 = ResourceBin(physical_units=self.units, term=self.term)
        source_list = [bin0.get_guid()]
        bin_list = [bin0]
        d0 = self.factory.make_delegation(units=self.units, term=self.term, rtype=self.type, sources=source_list,
                                          bins=bin_list, holder=ID(uid=self.factory.get_actor().get_name()))
        ticket0 = self.factory.make_ticket(delegation=d0)

        print(self.factory.to_json(ticket=ticket0))




#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum
from typing import List

from fabric.actor.core.apis.i_mgmt_actor import IMgmtActor
from fabric.actor.core.util.id import ID


class Status(Enum):
    OK = 0
    NOTOK = 1
    NOTREADY = 3


class StatusChecker:

    def check(self, *, controller: IMgmtActor, rid, ok: List[ID], not_ok: List[ID]) -> Status:
        result = self.check_(controller=controller, rid=0)
        if result == Status.OK:
            ok.append(rid)
        elif result == Status.NOTOK:
            not_ok.append(rid)
        return result

    def check_(self, *, controller: IMgmtActor, rid) -> Status:
        raise NotImplementedError



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import queue
import threading
from datetime import datetime
from typing import TYPE_CHECKING

from fabric.actor.core.container.aevent_subscription import AEventSubscription
from fabric.actor.core.util.iterable_queue import IterableQueue

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_event import IEvent
    from fabric.actor.security.auth_token import AuthToken


class EventSubscription(AEventSubscription):
    MAX_QUEUE_SIZE = 1000
    # Max time between drain calls. The subscription is considered
    # abandoned if the time between drain calls exceeds this limit.
    MAX_IDLE_TIME = 5 * 60 * 1000 * 1000

    def __init__(self, *, token: AuthToken, filters: list):
        super().__init__(token=token, filters=filters)
        self.queue = queue.Queue(maxsize=self.MAX_QUEUE_SIZE)
        self.last_drain_time_stamp = datetime.utcnow().timestamp()
        self.lock = threading.Lock()

    def deliver_event(self, *, event: IEvent):
        if not self.matches(event=event):
            return
        try:
            self.lock.acquire()
            if self.queue.qsize() == self.MAX_QUEUE_SIZE:
                self.queue.get()

            self.queue.put_nowait(event)
            self.queue.task_done()
        finally:
            self.lock.release()

    def drain_events(self, *, timeout: int) -> list:
        self.last_drain_time_stamp = datetime.utcnow().timestamp()
        result = []
        try:
            self.lock.acquire()
            for item in IterableQueue(source_queue=self.queue):
                result.append(item)
        finally:
            self.lock.release()

        return result

    def is_abandoned(self) -> bool:
        return datetime.utcnow().timestamp() - self.last_drain_time_stamp > self.MAX_IDLE_TIME



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

import threading

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import SliceException

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_kernel_slice import IKernelSlice
    from fabric.actor.core.util.id import ID


class SliceTable:
    def __init__(self):
        self.lock = threading.Lock()
        self.slices = {}
        self.inventory_slices = {}
        self.client_slices = {}
        self.broker_client_slices = {}
        # A map of sets of slices indexed by name. Name is not guaranteed
        # to be unique per slice.
        self.slices_by_name = {}

    def add(self, *, slice_object: IKernelSlice):
        """
        Adds the given slice to the slice table.

        @param slice_object slice to add

        @throws Exception if the slice is invalid
        @throws Exception if the slice is already present in the table
        """
        if slice_object.get_slice_id() is None or slice_object.get_name() is None:
            raise SliceException(Constants.invalid_argument)

        try:
            self.lock.acquire()

            if slice_object.get_slice_id() in self.slices:
                raise SliceException("Already exists")

            if slice_object.is_inventory():
                self.inventory_slices[slice_object.get_slice_id()] = slice_object
            elif slice_object.is_broker_client():
                self.broker_client_slices[slice_object.get_slice_id()] = slice_object
            elif slice_object.is_client():
                self.client_slices[slice_object.get_slice_id()] = slice_object
            else:
                raise SliceException("Unsupported slice type")

            self.slices[slice_object.get_slice_id()] = slice_object

            temp_dict = None
            if slice_object.get_name() in self.slices_by_name:
                temp_dict = self.slices_by_name[slice_object.get_name()]
                temp_dict[slice_object.get_slice_id()] = slice_object
            else:
                temp_dict = {slice_object.get_slice_id(): slice_object}

            self.slices_by_name[slice_object.get_name()] = temp_dict
        finally:
            self.lock.release()

    def contains(self, *, slice_id: ID) -> bool:
        """
        Checks if the specified slice is contained in the table.

        @param slice_id identifier of slice to check

        @return true if the slice is contained in the table, false otherwise
        """
        ret_val = False
        try:
            self.lock.acquire()
            ret_val = slice_id in self.slices.keys()
        finally:
            self.lock.release()
        return ret_val

    def get(self, *, slice_id: ID, raise_exception: bool = False) -> IKernelSlice:
        """
        Returns the specified slice.

        @param slice_id identifier of slice to return
        @param raise_exception: raise_exception

        @return slice or null if the slice is not present in the table
        """
        if slice_id is None:
            raise SliceException("Invalid argument")

        result = None
        try:
            self.lock.acquire()

            if slice_id in self.slices:
                result = self.slices[slice_id]
        finally:
            self.lock.release()
        if raise_exception and result is None:
            raise SliceException("not registered")
        return result

    def get_slice_name(self, *, slice_name: str) -> list:
        """
        Returns all slices with the given name.

        @param slice_name slice name

        @return an list of slices with the given name
        """
        ret_val = None
        try:
            self.lock.acquire()
            if slice_name in self.slices_by_name:
                temp_dict = self.slices_by_name[slice_name]
                ret_val = []
                for v in temp_dict.values():
                    ret_val.append(v)
        finally:
            self.lock.release()
        return ret_val

    def get_broker_client_slices(self) -> list:
        """
        Returns all broker client slices in the table.

        @return a list of broker client slices
        """
        ret_val = None
        try:
            self.lock.acquire()
            ret_val = []
            for v in self.broker_client_slices.values():
                ret_val.append(v)
        finally:
            self.lock.release()

        return ret_val

    def get_client_slices(self) -> list:
        """
        Returns all client slices in the table.

        @return an array of client slices
        """
        ret_val = None
        try:
            self.lock.acquire()
            ret_val = []
            for v in self.client_slices.values():
                ret_val.append(v)
            for v in self.broker_client_slices.values():
                ret_val.append(v)
        finally:
            self.lock.release()
        return ret_val

    def get_inventory_slices(self) -> list:
        """
        Returns all inventory slices in the table.

        @return a list of inventory slices
        """
        ret_val = None
        try:
            self.lock.acquire()
            ret_val = []
            for v in self.inventory_slices.values():
                ret_val.append(v)
        finally:
            self.lock.release()
        return ret_val

    def get_slices(self) -> list:
        """
        Returns all slices in the table.

        @return a list of slices
        """
        ret_val = None
        try:
            self.lock.acquire()
            ret_val = []
            for v in self.slices.values():
                ret_val.append(v)
        finally:
            self.lock.release()
        return ret_val

    def remove(self, *, slice_id: ID):
        """
        Removes the specified slice.

        @param slice_id identifier of slice to remove

        @throws Exception if the specified slice is not in the table
        """
        if slice_id is None:
            raise SliceException("Invalid argument")

        try:
            self.lock.acquire()
            slice_obj = self.slices.pop(slice_id)

            if slice_obj is None:
                raise SliceException("not registered")

            if slice_obj.is_inventory():
                self.inventory_slices.pop(slice_id)
            elif slice_obj.is_broker_client():
                self.broker_client_slices.pop(slice_id)
            elif slice_obj.is_client():
                self.client_slices.pop(slice_id)
            else:
                raise SliceException("Unsupported slice type")
            if slice_obj.get_name() in self.slices_by_name:
                entry = self.slices_by_name[slice_obj.get_name()]
                if entry is not None:
                    entry.pop(slice_obj.get_slice_id())
                    if len(entry) == 0:
                        self.slices_by_name.pop(slice_obj.get_name())
                    else:
                        self.slices_by_name[slice_obj.get_name()] = entry
        finally:
            self.lock.release()



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)

from sqlalchemy import JSON, ForeignKey, LargeBinary
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, String, Integer, Sequence

Base = declarative_base()


class Actors(Base):
    """
    Represents Actors Database Table
    """
    __tablename__ = 'Actors'
    act_id = Column(Integer, Sequence('act_id', start=1, increment=1), autoincrement=True, primary_key=True)
    act_name = Column(String, unique=True, nullable=False)
    act_guid = Column(String, nullable=False)
    act_type = Column(Integer, nullable=False)
    properties = Column(LargeBinary)


class Clients(Base):
    """
    Represents Clients Database Table
    """
    __tablename__ = 'Clients'
    clt_id = Column(Integer, Sequence('clt_id', start=1, increment=1), autoincrement=True, primary_key=True)
    clt_act_id = Column(Integer, ForeignKey('Actors.act_id'))
    clt_name = Column(String, nullable=False)
    clt_guid = Column(String, nullable=False)
    properties = Column(LargeBinary)


class ConfigMappings(Base):
    """
    Represents ConfigMappings Database Table
    """
    __tablename__ = 'ConfigMappings'
    cfgm_id = Column(Integer, Sequence('cfgm_id', start=1, increment=1), autoincrement=True, primary_key=True)
    cfgm_act_id = Column(Integer, ForeignKey('Actors.act_id'))
    cfgm_type = Column(String, nullable=False)
    cfgm_path = Column(String, nullable=False)
    properties = Column(JSON)


class ManagerObjects(Base):
    """
    Represents ManagerObjects Database Table
    """
    __tablename__ = 'ManagerObjects'
    mo_id = Column(Integer, Sequence('mo_id', start=1, increment=1), autoincrement=True, primary_key=True)
    mo_key = Column(String, nullable=False, unique=True)
    mo_act_id = Column(Integer, ForeignKey('Actors.act_id'))
    properties = Column(JSON)


class Miscellaneous(Base):
    """
    Represents Miscellaneous Database Table
    """
    __tablename__ = 'Miscellaneous'
    msc_id = Column(Integer, Sequence('msc_id', start=1, increment=1), autoincrement=True, primary_key=True)
    msc_path = Column(String, nullable=False, unique=True)
    properties = Column(JSON)


class Proxies(Base):
    """
    Represents Proxies Database Table
    """
    __tablename__ = 'Proxies'
    prx_id = Column(Integer, Sequence('prx_id', start=1, increment=1), autoincrement=True, primary_key=True)
    prx_act_id = Column(Integer, ForeignKey('Actors.act_id'))
    prx_name = Column(String)
    properties = Column(LargeBinary)


class Reservations(Base):
    """
    Represents Reservations Database Table
    """
    __tablename__ = 'Reservations'
    rsv_id = Column(Integer, Sequence('rsv_id', start=1, increment=1), autoincrement=True, primary_key=True)
    rsv_graph_id = Column(String, nullable=True)
    rsv_slc_id = Column(Integer, ForeignKey('Slices.slc_id'))
    rsv_resid = Column(String, nullable=False)
    rsv_category = Column(Integer, nullable=False)
    rsv_state = Column(Integer, nullable=False)
    rsv_pending = Column(Integer, nullable=False)
    rsv_joining = Column(Integer, nullable=False)
    properties = Column(LargeBinary)


class Slices(Base):
    """
    Represents Slices Database Table
    """
    __tablename__ = 'Slices'
    slc_id = Column(Integer, Sequence('slc_id', start=1, increment=1), autoincrement=True, primary_key=True)
    slc_graph_id = Column(String, nullable=True)
    slc_guid = Column(String, nullable=False)
    slc_name = Column(String, nullable=False)
    slc_type = Column(Integer, nullable=False)
    slc_act_id = Column(Integer, ForeignKey('Actors.act_id'))
    slc_resource_type = Column(String)
    properties = Column(LargeBinary)


class Units(Base):
    """
    Represents Units Database Table
    """
    __tablename__ = 'Units'
    unt_id = Column(Integer, Sequence('unt_id', start=1, increment=1), autoincrement=True, primary_key=True)
    unt_uid = Column(String)
    unt_unt_id = Column(Integer, nullable=True)
    unt_act_id = Column(Integer, ForeignKey('Actors.act_id'))
    unt_slc_id = Column(Integer, ForeignKey('Slices.slc_id'))
    unt_rsv_id = Column(Integer, ForeignKey('Reservations.rsv_id'))
    unt_type = Column(Integer, nullable=False)
    unt_state = Column(Integer, nullable=False)
    properties = Column(LargeBinary)


class Plugins(Base):
    """
    Represents Plugins Database Table
    """
    __tablename__ = 'Plugins'
    plg_id = Column(Integer, Sequence('plg_id', start=1, increment=1), autoincrement=True, primary_key=True)
    plg_local_id = Column(String, nullable=False)
    plg_type = Column(Integer, nullable=False)
    plg_actor_type = Column(Integer)
    properties = Column(LargeBinary)


class Delegations(Base):
    """
    Represents Delegations Database Table
    """
    __tablename__ = 'Delegations'
    dlg_id = Column(Integer, Sequence('dlg_id', start=1, increment=1), autoincrement=True, primary_key=True)
    dlg_slc_id = Column(Integer, ForeignKey('Slices.slc_id'))
    dlg_act_id = Column(Integer, ForeignKey('Actors.act_id'))
    dlg_graph_id = Column(String, nullable=False)
    dlg_state = Column(Integer, nullable=False)
    properties = Column(LargeBinary)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import unittest
from datetime import datetime, time

from fabric.actor.core.apis.i_database import IDatabase
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.core.unit import Unit
from fabric.actor.core.kernel.authority_reservation_factory import AuthorityReservationFactory
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.plugins.substrate.db.substrate_actor_database import SubstrateActorDatabase
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.test.core.plugins.db.actor_database_test import ActorDatabaseTest


class SubstrateDatabaseTest(ActorDatabaseTest):

    logger = logging.getLogger('AuthorityPolicyTest')
    log_format = '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(format=log_format, filename="actor.log")
    Logger.setLevel(logging.INFO)

    def make_actor_database(self) -> IDatabase:
        db = SubstrateActorDatabase(user=self.db_user, password=self.db_pwd, database=self.db_name, db_host=self.db_host,
                                    logger=self.logger)
        return db

    def test_d_add_update_get_unit(self):
        actor = self.prepare_actor_database()
        db = actor.get_plugin().get_database()
        slice_obj = SliceFactory.create(slice_id=ID(), name="slice-1")
        db.add_slice(slice_object=slice_obj)

        rset = ResourceSet()
        term = Term(start=datetime.now(), end=datetime.now().replace(minute=20))
        res = AuthorityReservationFactory.create(resources=rset, term=term, slice_obj=slice_obj, rid=ID())
        db.add_reservation(reservation=res)

        rtype = ResourceType(resource_type="12")
        u = Unit(id=ID(), rid=res.get_reservation_id(), slice_id=slice_obj.get_slice_id(), actor_id=actor.get_guid())
        u.set_resource_type(rtype=rtype)
        db.add_unit(u=u)

        self.assertIsNotNone(db.get_unit(unit_id=u.get_id()))



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import pickle
from enum import Enum


from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import UnitException
from fabric.actor.core.plugins.config.config_token import ConfigToken
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.notice import Notice
from fabric.actor.core.util.resource_type import ResourceType


class UnitState(Enum):
    DEFAULT = 0
    PRIMING = 1
    ACTIVE = 2
    MODIFYING = 3
    CLOSING = 4
    CLOSED = 5
    FAILED = 6


class Unit(ConfigToken):
    def __init__(self, *, id: ID, rid: ID = None, slice_id: ID = None, actor_id: ID = None,
                 properties: dict = None, state: UnitState = None):
        # Unique identifier.
        self.id = id
        # Resource type.
        self.rtype = None
        # Unique identifier of parent unit (optional).
        self.parent_id = None
        # Properties list.
        self.properties = {}
        if properties is not None:
            self.properties = properties
        # Unit state
        self.state = UnitState.DEFAULT
        if state is not None:
            self.state = state
        # Configuration sequence number. Each unique configuration action is identified by a sequence number.
        self.sequence = 0
        # Reservation this unit belongs to (id).
        self.reservation_id = rid
        # Slice this unit belongs to.
        self.slice_id = slice_id
        # Actor this unit belongs to.
        self.actor_id = actor_id
        self.notices = Notice()
        # Reservation this unit belongs to.
        self.reservation = None
        # The modified version of this unit.
        self.modified = None
        self.transfer_out_started = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['transfer_out_started']
        del state['modified']
        del state['reservation']

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.transfer_out_started = False
        self.modified = False
        self.reservation = None

    def transition(self, *, to_state: UnitState):
        """
        Transition the state
        @param to_state ti state
        """
        self.state = to_state

    def merge_properties(self, *, incoming: dict):
        """
        Merge properties
        @param incoming incoming properties
        """
        self.properties = {**incoming, **self.properties}

    def fail(self, *, message: str, exception: Exception = None):
        """
        Fail a unit
        @param message
        @param exception exception
        """
        self.notices.add(msg=message, ex=exception)
        self.transition(to_state=UnitState.FAILED)

    def fail_on_modify(self, *, message: str, exception: Exception = None):
        """
        Fail on modify
        @param message message
        @param exception exception
        """
        self.notices.add(msg=message, ex=exception)
        self.transition(to_state=UnitState.ACTIVE)
        self.merge_properties(incoming=self.modified.properties)

    def set_state(self, *, state: UnitState):
        """
        Set state
        @param state state
        """
        self.transition(to_state=state)

    def start_close(self):
        """
        Start close on a unit
        """
        self.transition(to_state=UnitState.CLOSING)
        self.transfer_out_started = True

    def start_prime(self) -> bool:
        """
        Start priming the unit
        """
        if self.state == UnitState.DEFAULT or self.state == UnitState.PRIMING:
            self.transition(to_state=UnitState.PRIMING)
            return True
        return False

    def start_modify(self) -> bool:
        """
        Start modifying
        """
        if self.state == UnitState.ACTIVE or self.state == UnitState.MODIFYING:
            self.transition(to_state=UnitState.MODIFYING)
            return True
        return False

    def activate(self):
        """
        Mark the unit as active
        """
        self.state = UnitState.ACTIVE

    def close(self):
        """
        Mark the unit as closed
        """
        self.state = UnitState.CLOSED

    def get_id(self) -> ID:
        """
        Get unit id
        @return unit id
        """
        return self.id

    def get_properties(self) -> dict:
        """
        Get unit properties
        @return unit properties
        """
        return self.properties

    def set_property(self, *, name: str, value: str):
        """
        Set property
        @param name name
        @param value value
        """
        self.properties[name] = value

    def get_property(self, *, name: str) -> str:
        """
        Get Property
        @param name name
        @return value
        """
        ret_val = None
        if name in self.properties:
            ret_val = self.properties[name]
        return ret_val

    def get_state(self) -> UnitState:
        """
        Get Unit state
        @return unit state
        """
        return self.state

    def clone(self):
        """
        Clone a unit
        @return copy of the object
        """
        ret_val = Unit(id=self.id, properties=self.properties, state=self.state)
        return ret_val

    def get_sequence(self) -> int:
        """
        Get Sequence number
        @return sequence number
        """
        return self.sequence

    def get_sequence_increment(self) -> int:
        """
        Get Sequence number and increment it
        @return sequence number and increment it
        """
        self.sequence += 1
        ret_val = self.sequence
        return ret_val

    def increment_sequence(self) -> int:
        """
        Increment sequence number
        """
        self.sequence += 1
        ret_val = self.sequence
        return ret_val

    def decrement_sequence(self) -> int:
        """
        Decrement sequence number
        """
        self.sequence -= 1
        ret_val = self.sequence
        return ret_val

    def set_reservation(self, *, reservation: IReservation):
        """
        Set Reservation
        @param reservation reservation
        """
        self.reservation = reservation
        self.reservation_id = reservation.get_reservation_id()

    def set_slice_id(self, *, slice_id: ID):
        """
        Set slice id
        @param slice_id slice id
        """
        self.slice_id = slice_id

    def set_actor_id(self, *, actor_id: ID):
        """
        Set actor id
        @param actor_id actor id
        """
        self.actor_id = actor_id

    def is_failed(self) -> bool:
        """
        Check fail status
        @return true if unit is failed, false otherwise
        """
        ret_val = self.state == UnitState.FAILED
        return ret_val

    def is_closed(self) -> bool:
        """
        Check close status
        @return true if unit is closed, false otherwise
        """
        ret_val = self.state == UnitState.CLOSED
        return ret_val

    def is_active(self) -> bool:
        """
        Check active status
        @return true if unit is active, false otherwise
        """
        ret_val = self.state == UnitState.ACTIVE
        return ret_val

    def has_pending_action(self) -> bool:
        """
        Check pending action status
        @return true if unit has pending operation, false otherwise
        """
        return self.state == UnitState.MODIFYING or self.state == UnitState.PRIMING or self.state == UnitState.CLOSING

    def set_modified(self, *, modified):
        """
        Set modified unit
        @param modified modified
        """
        self.modified = modified

    def get_modified(self):
        """
        Get modified unit
        @return modified unit
        """
        return self.modified

    def get_slice_id(self) -> ID:
        """
        Return slice id
        @return slice id
        """
        return self.slice_id

    def get_reservation_id(self) -> ID:
        """
        Get Reservation Id
        @return reservation id
        """
        return self.reservation_id

    def get_reservation(self) -> IReservation:
        """
        Get Reservation
        @return reservation
        """
        ret_val = self.reservation
        return ret_val

    def get_parent_id(self) -> ID:
        """
        Get Parent Id
        @return parent guid
        """
        return self.parent_id

    def set_parent_id(self, *, parent_id: ID):
        """
        Set Parent id
        @param parent_id parent id
        """
        self.parent_id = parent_id

    def complete_modify(self):
        """
        Complete Modify operation
        """
        self.transition(to_state=UnitState.ACTIVE)
        self.merge_properties(incoming=self.modified.properties)

    def get_actor_id(self) -> ID:
        """
        Get Actor id
        @return actor id
        """
        return self.actor_id

    def get_resource_type(self) -> ResourceType:
        """
        Get Resource Type
        @return resource type
        """
        return self.rtype

    def set_resource_type(self, *, rtype: ResourceType):
        """
        Set resource type
        @param rtype resource type
        """
        self.rtype = rtype

    def __eq__(self, other):
        if not isinstance(other, Unit):
            return NotImplemented

        return self.id == other.id

    def get_notices(self) -> Notice:
        """
        Get notices
        @return notices
        """
        ret_val = self.notices
        return ret_val

    def add_notice(self, *, notice: str):
        """
        Add a notice
        @param notice notice
        """
        self.notices.add(msg=notice)

    def __str__(self):
        return "[unit: {} reservation: {} actor: {} state: {}]".format(self.id, self.reservation_id, self.actor_id, self.state)

    def __hash__(self):
        return self.id.__hash__()

    @staticmethod
    def create_instance(properties: dict):
        """
        Create an Unit instance using the pickled instance read from the database
        @param properties properties
        """
        if Constants.property_pickle_properties not in properties:
            raise UnitException(Constants.invalid_argument)
        serialized_unit = properties[Constants.property_pickle_properties]
        deserialized_unit = pickle.loads(serialized_unit)
        return deserialized_unit



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from enum import Enum
from typing import TYPE_CHECKING

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ResourcesException
from fabric.actor.core.kernel.slice_factory import SliceFactory

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor_identity import IActorIdentity
    from fabric.actor.core.apis.i_database import IDatabase
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.util.resource_data import ResourceData
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.actor.core.apis.i_slice import ISlice


class PoolManagerError(Enum):
    """
    Enumeration for Pool Manager errors
    """
    ErrorNone = 0
    ErrorPoolExists = -10
    ErrorTypeExists = -20
    ErrorInvalidArguments = -30
    ErrorDatabaseError = -40
    ErrorInternalError = -50


class CreatePoolResult:
    """
    Result of Create Pool
    """
    def __init__(self):
        self.code = PoolManagerError.ErrorNone
        self.slice = None


class PoolManager:
    """
    Implements the class responsible for creating pools on startup
    """
    def __init__(self, *, db: IDatabase, identity: IActorIdentity, logger):
        if db is None or identity is None or logger is None:
            raise ResourcesException("Invalid arguments {} {} {}".format(db, identity, logger))
        self.db = db
        self.identity = identity
        self.logger = logger

    def create_pool(self, *, slice_id: ID, name: str, rtype: ResourceType,
                    resource_data: ResourceData) -> CreatePoolResult:
        """
        Create Inventory Pool at boot
        @param slice_id slice id
        @param name name
        @param rtype resource type
        @param resource_data resource data
        """
        result = CreatePoolResult()
        if slice_id is None or name is None or rtype is None:
            result.code = PoolManagerError.ErrorInvalidArguments
            return result
        try:
            temp = self.db.get_slice(slice_id=slice_id)

            if temp is not None and len(temp) > 0:
                result.code = PoolManagerError.ErrorPoolExists
                return result

            temp = self.db.get_inventory_slices()
            if temp is not None and len(temp) > 0:
                for properties in temp:
                    slice_obj = SliceFactory.create_instance(properties=properties)
                    rt = slice_obj.get_resource_type()
                    if rt == rtype:
                        result.slice = slice_obj
                        result.code = PoolManagerError.ErrorTypeExists
                        return result

            slice_obj = SliceFactory.create(slice_id=slice_id, name=name, data=resource_data.clone())
            slice_obj.set_inventory(value=True)
            slice_obj.set_owner(owner=self.identity.get_identity())
            slice_obj.set_resource_type(resource_type=rtype)

            try:
                self.db.add_slice(slice_object=slice_obj)
                result.slice = slice_obj
            except Exception:
                traceback.print_exc()
                result.code = PoolManagerError.ErrorDatabaseError
        except Exception:
            traceback.print_exc()
            result.code = PoolManagerError.ErrorInternalError
        return result

    def update_pool(self, *, slice_obj: ISlice):
        """
        Update the resource pool
        @param slice_obj slice object
        """
        try:
            self.db.update_slice(slice_object=slice_obj)
        except Exception as e:
            raise ResourcesException("Could not update slice {}".format(e))

    def remove_pool(self, *, pool_id: ID, rtype: ResourceType):
        """
        Remove a pool
        @param pool_id pool id
        @param rtype resource type
        """
        temp = self.db.get_slice(slice_id=pool_id)

        if temp is not None and len(temp) > 0:
            slice_obj = SliceFactory.create_instance(properties=temp)
            if not slice_obj.is_inventory() or rtype != slice_obj.get_resource_type():
                raise ResourcesException(Constants.invalid_argument)

            self.db.remove_slice(slice_id=pool_id)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
import logging
import traceback
import unittest
from datetime import datetime
import time

from fabric.actor.core.apis.i_authority_proxy import IAuthorityProxy
from fabric.actor.core.apis.i_broker import IBroker
from fabric.actor.core.apis.i_broker_reservation import IBrokerReservation
from fabric.actor.core.apis.i_reservation import IReservation
from fabric.actor.core.apis.i_slice import ISlice
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import BrokerException
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.proxies.kafka.kafka_authority_proxy import KafkaAuthorityProxy
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.time.actor_clock import ActorClock
from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.security.auth_token import AuthToken
from fabric.actor.test.base_test_case import BaseTestCase
from fabric.actor.test.client_callback_helper import ClientCallbackHelper
from fabric.actor.test.dummy_authority_proxy import DummyAuthorityProxy


class BrokerPolicyTest(BaseTestCase):
    DonateStartCycle = 10
    DonateEndCycle = 100
    logger = logging.getLogger('BrokerPolicyTest')
    log_format = '%(asctime)s - %(name)s - {%(filename)s:%(lineno)d} - [%(threadName)s] - %(levelname)s - %(message)s'
    logging.basicConfig(format=log_format, filename="actor.log")
    logger.setLevel(logging.INFO)

    def get_broker(self, *, name: str = BaseTestCase.broker_name, guid: ID = BaseTestCase.broker_guid) -> IBroker:
        db = self.get_container_database()
        db.reset_db()
        broker = super().get_broker()
        broker.set_recovered(value=True)
        Term.set_clock(broker.get_actor_clock())
        return broker

    def get_source(self, units: int, rtype: ResourceType, broker: IBroker, slice_obj: ISlice):
        raise BrokerException(Constants.not_implemented)

    def get_request(self, units: int, rtype: ResourceType, start: datetime, end: datetime):
        raise BrokerException(Constants.not_implemented)

    def get_request_from_request(self, request: IBrokerReservation, units: int, rtype: ResourceType, start: datetime, end: datetime):
        raise BrokerException(Constants.not_implemented)

    def assert_ticketed(self, r: IReservation, units: int, rtype: ResourceType, start: datetime, end: datetime):
        self.assertIsNotNone(r)
        self.assertFalse(r.is_failed())
        self.assertEqual(units, r.get_resources().get_units())
        self.assertEqual(rtype, r.get_resources().get_type())
        self.assertIsNotNone(r.get_term())
        self.assertEqual(start, r.get_term().get_new_start_time())
        self.assertEqual(end, r.get_term().get_end_time())

    def get_authority_proxy(self) -> IAuthorityProxy:
        auth = AuthToken(name="mysite", guid=ID())
        from fabric.actor.core.container.globals import GlobalsSingleton
        proxy = KafkaAuthorityProxy(kafka_topic="test-topic", identity=auth, logger=GlobalsSingleton.get().get_logger())

        if ActorRegistrySingleton.get().get_proxy(Constants.protocol_local, "mysite") is None:
            dummy = DummyAuthorityProxy(auth=auth)
            ActorRegistrySingleton.get().register_proxy(dummy)
        return proxy

    def _test_a_create(self):
        """
        Tests if the actor and the policy can be instantiated.
        """
        broker = self.get_broker()
        self.assertIsNotNone(broker)

    def _test_b_allocate_ticket(self):
        """
        Requests a ticket for all resources. Checks if the ticket is
        allocated for what was asked. Checks the term. Checks whether the
        reservation is closed when it expires.
        """
        broker = self.get_broker()
        controller = self.get_controller()
        clock = broker.get_actor_clock()
        rtype = ResourceType(resource_type="1")

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(proxy)
        ActorRegistrySingleton.get().register_callback(broker_callback)
        last_called = proxy.get_called()

        inv_slice = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        inv_slice.set_inventory(value=True)
        source = self.get_source(1, rtype, broker, inv_slice)

        broker.donate_reservation(reservation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)

        request = self.get_request(units, rtype, start, end)

        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            if last_called < proxy.get_called():
                self.assert_ticketed(proxy.get_reservation(), units, rtype, start, end)
                last_called = proxy.get_called()

        broker.await_no_pending_reservations()

        self.assertEqual(1, proxy.get_called())
        self.assertTrue(request.is_closed())

    def _test_c_allocate_ticket2(self):
        """
        Requests a ticket for all resources. Checks if the ticket is
        allocated for what was asked. Checks the term. Checks whether the
        reservation is closed when it expires. Repeat one more time.
        """
        broker = self.get_broker()
        controller = self.get_controller()
        clock = broker.get_actor_clock()
        rtype = ResourceType(resource_type="1")

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(proxy)
        ActorRegistrySingleton.get().register_callback(broker_callback)
        last_called = proxy.get_called()

        inv_slice = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        inv_slice.set_inventory(value=True)
        source = self.get_source(1, rtype, broker, inv_slice)

        broker.donate_reservation(reservation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        cycle_end = self.DonateEndCycle - 50
        end = clock.cycle_end_date(cycle=cycle_end)

        request = self.get_request(units, rtype, start, end)

        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            if last_called < proxy.get_called():
                self.assert_ticketed(proxy.get_reservation(), units, rtype, start, end)
                last_called = proxy.get_called()

            if c == cycle_end:
                broker.await_no_pending_reservations()
                self.assertTrue(request.is_closed())
                self.assertEqual(1, proxy.get_called())
                start = clock.cycle_start_date(cycle=cycle_end + 1)
                end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)
                request = self.get_request(units, rtype, start, end)
                broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        broker.await_no_pending_reservations()

        self.assertEqual(2, proxy.get_called())
        self.assertTrue(request.is_closed())

    def test_d_extend_ticket(self):
        """
        Requests a ticket for all resources. Checks if the ticket is
        allocated for what was asked. Checks the term. Checks whether the
        reservation is closed when it expires. Repeat one more time.
        """
        broker = self.get_broker()
        controller = self.get_controller()
        clock = broker.get_actor_clock()
        rtype = ResourceType(resource_type="1")

        proxy = ClientCallbackHelper(name=controller.get_name(), guid=controller.get_guid())
        broker_callback = ClientCallbackHelper(name=broker.get_name(), guid=broker.get_guid())
        ActorRegistrySingleton.get().register_callback(callback=proxy)
        ActorRegistrySingleton.get().register_callback(callback=broker_callback)
        last_called = proxy.get_called()

        inv_slice = SliceFactory.create(slice_id=ID(), name="inventory-slice")
        inv_slice.set_inventory(value=True)
        source = self.get_source(1, rtype, broker, inv_slice)

        broker.donate_reservation(reservation=source)

        cycle = 1
        broker.external_tick(cycle=cycle)
        cycle += 1

        units = 1
        start = clock.cycle_start_date(cycle=self.DonateStartCycle)
        cycle_end = self.DonateEndCycle - 50
        end = clock.cycle_end_date(cycle=cycle_end)

        request = self.get_request(units, rtype, start, end)
        reservation = request

        broker.ticket(reservation=request, callback=proxy, caller=proxy.get_identity())

        for c in range(cycle, self.DonateEndCycle):
            broker.external_tick(cycle=c)
            while broker.get_current_cycle() != c:
                time.sleep(0.001)

            if last_called < proxy.get_called():
                self.assert_ticketed(proxy.get_reservation(), units, rtype, start, end)
                last_called = proxy.get_called()

            if c == cycle_end - 10:
                self.assertEqual(1, proxy.get_called())
                start = ActorClock.from_milliseconds(milli_seconds=ActorClock.to_milliseconds(when=end) + 1)
                end = clock.cycle_end_date(cycle=self.DonateEndCycle - 1)
                request = self.get_request_from_request(request, units, rtype, start, end)
                broker.extend_ticket(reservation=request, caller=proxy.get_identity())
                print("Extend done")

        broker.await_no_pending_reservations()

        self.assertEqual(2, proxy.get_called())
        self.assertTrue(reservation.is_closed())


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

import traceback
from typing import TYPE_CHECKING, List

from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ManageException
from fabric.actor.core.manage.error import Error
from fabric.actor.core.apis.i_component import IComponent
from fabric.actor.core.manage.kafka.kafka_mgmt_message_processor import KafkaMgmtMessageProcessor
from fabric.actor.core.manage.messages.protocol_proxy_mng import ProtocolProxyMng
from fabric.message_bus.messages.result_avro import ResultAvro

if TYPE_CHECKING:
    from fabric.actor.core.util.id import ID
    from fabric.message_bus.messages.auth_avro import AuthAvro
    from fabric.message_bus.producer import AvroProducerApi


class KafkaProxy(IComponent):
    def __init__(self, *, guid: ID, kafka_topic: str, auth: AuthAvro, logger,
                 message_processor: KafkaMgmtMessageProcessor, producer: AvroProducerApi = None):
        self.management_id = guid
        self.auth = auth
        self.logger = logger
        self.kafka_topic = kafka_topic
        self.producer = None
        self.last_status = None
        self.last_exception = None
        self.callback_topic = None
        if producer is None:
            self.setup_kafka_producer()
        else:
            self.producer = producer
        self.message_processor = message_processor

    def setup_kafka_producer(self):
        try:
            from fabric.actor.core.container.globals import GlobalsSingleton
            self.producer = GlobalsSingleton.get().get_kafka_producer()
        except Exception as e:
            traceback.print_exc()
            self.logger.eror("Exception occurred while loading schemas {}".format(e))
            raise e

    def get_kafka_topic(self):
        return self.kafka_topic

    def clear_last(self):
        self.last_exception = None
        self.last_status = ResultAvro()

    def get_last_error(self) -> Error:
        return Error(status=self.last_status, e=self.last_exception)

    def get_protocols(self) -> List[ProtocolProxyMng]:
        proto = ProtocolProxyMng()
        proto.set_protocol(protocol=Constants.protocol_kafka)
        proto.set_proxy_class(proxy_class=KafkaProxy.__class__.__name__)
        proto.set_proxy_module(proxy_module=KafkaProxy.__module__)
        result = [proto]
        return result

    def get_type_id(self) -> str:
        raise ManageException("Not implemented")


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from datetime import datetime
from typing import List

from fabric.actor.core.common.constants import Constants

"""
Implements various classes for reading and loading the information from config file
"""


class GlobalConfig:
    """
    Represent Global configuration
    """
    def __init__(self, *, config: dict):
        self.runtime = {}
        if Constants.config_section_runtime in config:
            for prop in config[Constants.config_section_runtime]:
                for key, value in prop.items():
                    self.runtime[key] = value
        self.logging = {}
        if Constants.config_logging_section in config:
            for prop in config[Constants.config_logging_section]:
                for key, value in prop.items():
                    self.logging[key] = value
        self.oauth = {}
        if Constants.config_section_o_auth in config:
            for prop in config[Constants.config_section_o_auth]:
                for key, value in prop.items():
                    self.oauth[key] = value
        self.database = {}
        if Constants.config_section_database in config:
            for prop in config[Constants.config_section_database]:
                for key, value in prop.items():
                    self.database[key] = value
        self.container = {}
        if Constants.config_section_container in config:
            for prop in config[Constants.config_section_container]:
                for key, value in prop.items():
                    self.container[key] = value
        self.time = {}
        if Constants.config_section_time in config:
            for prop in config[Constants.config_section_time]:
                for key, value in prop.items():
                    self.time[key] = value

        self.neo4j = {}
        if Constants.config_section_neo4j in config:
            self.neo4j = config[Constants.config_section_neo4j]

        self.pdp = {}
        if Constants.config_section_pdp in config:
            self.pdp = config[Constants.config_section_pdp]

    def get_runtime(self) -> dict:
        """
        Return runtime config
        """
        return self.runtime

    def get_logging(self) -> dict:
        """
        Return logging config
        """
        return self.logging

    def get_oauth(self) -> dict:
        """
        Return oauth config
        """
        return self.oauth

    def get_database(self) -> dict:
        """
        Return database config
        """
        return self.database

    def get_container(self) -> dict:
        """
        Return container config
        """
        return self.container

    def get_time(self) -> dict:
        """
        Return time config
        """
        return self.time

    def get_neo4j_config(self) -> dict:
        """
        Return neo4j config
        """
        return self.neo4j

    def get_pdp_config(self) -> dict:
        """
        Return PDP config
        """
        return self.pdp


class HandlerConfig:
    """
    Represents Handler Config
    """
    def __init__(self, *, config: list):
        self.module_name = None
        self.class_name = None
        self.properties = {}
        for prop in config:
            for key, value in prop.items():
                if key == 'module':
                    self.module_name = value
                elif key == 'class_name':
                    self.class_name = value
                elif key == 'properties':
                    for p in value:
                        for k, v in p.items():
                            self.properties[k] = str(v)

    def get_module_name(self) -> str:
        """
        Return Module Name for Handler
        """
        return self.module_name

    def get_class_name(self) -> str:
        """
        Return Class Name for Handler
        """
        return self.class_name

    def get_properties(self) -> dict:
        """
        Return Handler Properties
        """
        return self.properties


class Attribute:
    """
    Represents Attributes
    """
    def __init__(self, *, config: list):
        self.key = None
        self.type = None
        self.value = None
        for prop in config:
            for key, value in prop.items():
                if key == 'key':
                    self.key = value
                elif key == 'type':
                    self.type = value
                elif key == 'value':
                    self.value = value

    def get_type(self) -> str:
        """
        Return Attribute Type
        """
        return self.type

    def get_key(self) -> str:
        """
        Return Attribute Key
        """
        return self.key

    def get_value(self) -> str:
        """
        Return Attribute Value
        """
        return self.value


class PoolConfig:
    """
    Represents Pool Config
    """
    def __init__(self, *, config: list):
        self.description = None
        self.attributes = []
        self.type = None
        self.label = None
        self.units = None
        now = datetime.utcnow()
        self.start = now
        self.end = now.replace(year=now.year + 20)
        self.properties = {}
        self.factory_module = None
        self.factory_class = None
        self.handler = None
        for prop in config:
            for key, value in prop.items():
                if key == 'type':
                    self.type = value
                elif key == 'label':
                    self.label = value
                elif key == 'description':
                    self.description = value
                elif key == 'units':
                    self.units = int(value)
                elif key == 'start':
                    self.start = value
                elif key == 'end':
                    self.end = value
                elif key == 'properties':
                    for p in value:
                        for k, v in p.items():
                            self.properties[k] = str(v)
                elif key == 'factory.class':
                    self.factory_class = value
                elif key == 'factory.module':
                    self.factory_module = value
                elif key == 'handler':
                    self.handler = HandlerConfig(config=value)
                elif key == 'attributes':
                    for a in value:
                        self.attributes.append(Attribute(config=a['attribute']))

    def get_type(self) -> str:
        """
        Return Pool Type
        """
        return self.type

    def get_label(self) -> str:
        """
        Return Pool Label
        """
        return self.label

    def get_description(self) -> str:
        """
        Return Pool Description
        """
        return self.description

    def get_units(self) -> int:
        """
        Return Pool Units
        """
        return self.units

    def get_start(self) -> datetime:
        """
        Return Start lease time
        """
        return self.start

    def get_end(self) -> datetime:
        """
        Return End lease time
        """
        return self.end

    def get_handler(self) -> HandlerConfig:
        """
        Return Handler Config
        """
        return self.handler

    def get_properties(self) -> dict:
        """
        Return Properties
        """
        return self.properties

    def get_factory_class(self) -> str:
        """
        Return Factory class
        """
        return self.factory_class

    def get_factory_module(self) -> str:
        """
        Return Factory Module
        """
        return self.factory_module

    def get_attributes(self) -> List[Attribute]:
        """
        Return Attributes
        """
        return self.attributes


class ControlConfig:
    """
    Represents Control config
    """
    def __init__(self, *, config: list):
        self.type = None
        self.module_name = None
        self.class_name = None
        for prop in config:
            for key, value in prop.items():
                if key == 'type':
                    self.type = value
                elif key == 'module':
                    self.module_name = value
                elif key == 'class':
                    self.class_name = value

    def get_type(self) -> str:
        """
        Return type
        """
        return self.type

    def get_module_name(self) -> str:
        """
        Return Module name
        """
        return self.module_name

    def get_class_name(self) -> str:
        """
        Return Class Name
        """
        return self.class_name


class ResourceConfig:
    """
    Represents resource config
    """
    def __init__(self, *, config: list):
        self.description = None
        self.attributes = []
        self.type = None
        self.label = None
        self.properties = {}
        self.resource_module = None
        self.resource_class = None
        self.handler = None
        for prop in config:
            for key, value in prop.items():
                if key == 'type':
                    self.type = value
                elif key == 'label':
                    self.label = value
                elif key == 'description':
                    self.description = value
                elif key == 'properties':
                    for p in value:
                        for k, v in p.items():
                            self.properties[k] = str(v)
                elif key == 'resource_class':
                    self.resource_class = value
                elif key == 'resource_module':
                    self.factory_module = value
                elif key == 'handler':
                    self.handler = HandlerConfig(config=value)
                elif key == 'control':
                    self.control = ControlConfig(config=value)
                elif key == 'attributes':
                    for a in value:
                        self.attributes.append(Attribute(config=a['attribute']))

    def get_type(self) -> str:
        """
        Return type
        """
        return self.type

    def get_label(self) -> str:
        """
        Return label
        """
        return self.label

    def get_description(self) -> str:
        """
        Return description
        """
        return self.description

    def get_handler(self) -> HandlerConfig:
        """
        Return handler config
        """
        return self.handler

    def get_control(self) -> ControlConfig:
        """
        Return control config
        """
        return self.control

    def get_properties(self) -> dict:
        """
        Return properties
        """
        return self.properties

    def get_resource_class(self) -> str:
        """
        Return resource class
        """
        return self.resource_class

    def get_resource_module(self) -> str:
        """
        Return resource module
        """
        return self.resource_module

    def get_attributes(self) -> List[Attribute]:
        """
        Return resource attributes
        """
        return self.attributes


class PolicyConfig:
    """
    Represents Policy config
    """
    def __init__(self, *, config: list):
        self.module_name = None
        self.class_name = None
        self.properties = {}
        for prop in config:
            for key, value in prop.items():
                if key == 'module':
                    self.module_name = value
                elif key == 'class':
                    self.class_name = value
                elif key == 'properties':
                    for p in value:
                        for k, v in p.items():
                            self.properties[k] = str(v)

    def get_module_name(self) -> str:
        """
        Return module name
        """
        return self.module_name

    def get_class_name(self) -> str:
        """
        Return class name
        """
        return self.class_name

    def get_properties(self) -> dict:
        """
        Return properties
        """
        return self.properties


class ActorConfig:
    """
    Represents Actor Config
    """
    def __init__(self, *, config: list):
        self.policy = None
        self.description = None
        self.pools = []
        self.resources = []
        self.controls = []
        self.type = None
        self.name = None
        self.guid = None
        self.kafka_topic = None
        self.policy = None
        self.substrate_file = None
        for prop in config:
            for key, value in prop.items():
                if key == 'type':
                    self.type = value
                elif key == 'name':
                    self.name = value
                elif key == 'guid':
                    self.guid = value
                elif key == 'description':
                    self.description = value
                elif key == 'kafka-topic':
                    self.kafka_topic = value
                elif key == 'substrate.file':
                    self.substrate_file = value
                elif key == 'pools':
                    for p in value:
                        self.pools.append(PoolConfig(config=p['pool']))
                elif key == 'controls':
                    for c in value:
                        self.controls.append(ControlConfig(config=c['control']))
                elif key == 'policy':
                    self.policy = PolicyConfig(config=value)
                elif key == 'resources':
                    for p in value:
                        self.resources.append(ResourceConfig(config=p['resource']))

    def get_type(self) -> str:
        """
        Return Actor Type
        """
        return self.type

    def get_name(self) -> str:
        """
        Return Actor Name
        """
        return self.name

    def get_guid(self) -> str:
        """
        Return Actor Guid
        """
        return self.guid

    def get_description(self) -> str:
        """
        Return Actor Description
        """
        return self.description

    def get_kafka_topic(self) -> str:
        """
        Return Kafka topic
        """
        return self.kafka_topic

    def get_pools(self) -> List[PoolConfig]:
        """
        Return Pools
        """
        return self.pools

    def get_controls(self) -> List[ControlConfig]:
        """
        Return Controls
        """
        return self.controls

    def get_policy(self) -> PolicyConfig:
        """
        Return Policy
        """
        return self.policy

    def get_resources(self) -> List[ResourceConfig]:
        """
        Return resources
        """
        return self.resources

    def get_substrate_file(self) -> str:
        """
        Return Substrate file
        """
        return self.substrate_file


class Peer:
    """
    Represent Peer Config
    """
    def __init__(self, *, config: list):
        self.name = None
        self.type = None
        self.guid = None
        self.kafka_topic = None
        self.delegation = None
        for prop in config:
            for key, value in prop.items():
                if key == 'name':
                    self.name = value
                elif key == 'type':
                    self.type = value
                elif key == 'guid':
                    self.guid = value
                elif key == 'kafka-topic':
                    self.kafka_topic = value
                elif key == 'delegation':
                    self.delegation = value

    def get_name(self) -> str:
        """
        Return Name
        """
        return self.name

    def get_type(self) -> str:
        """
        Return Type
        """
        return self.type

    def get_guid(self) -> str:
        """
        Return Guid
        """
        return self.guid

    def get_kafka_topic(self) -> str:
        """
        Return Kafka Topic
        """
        return self.kafka_topic

    def get_delegation(self) -> str:
        """
        Return delegation
        """
        return self.delegation


class Configuration:
    """
    Represent Configuration read from file
    """
    def __init__(self, *, config: dict):
        self.global_config = GlobalConfig(config=config)
        self.actor = ActorConfig(config=config['actor'])
        self.peers = []
        if 'peers' in config:
            for e in config['peers']:
                self.peers.append(Peer(config=e['peer']))

    def get_global_config(self) -> GlobalConfig:
        """
        Return Global Config
        """
        return self.global_config

    def get_runtime_config(self) -> dict:
        """
        Return Runtime Config
        """
        if self.global_config is not None:
            return self.global_config.get_runtime()
        return None

    def get_oauth_config(self) -> dict:
        """
        Return OAuth Config
        """
        if self.global_config is not None:
            return self.global_config.get_oauth()
        return None

    def get_actor(self) -> ActorConfig:
        """
        Return Actor Config
        """
        return self.actor

    def get_peers(self) -> List[Peer]:
        """
        Return Peer Config
        """
        return self.peers



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from typing import TYPE_CHECKING

from fabric.actor.core.apis.i_delegation import IDelegation
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ProxyException
from fabric.actor.core.kernel.resource_set import ResourceSet
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.time.actor_clock import ActorClock

from fabric.actor.core.time.term import Term
from fabric.actor.core.util.id import ID
from fabric.actor.core.util.prop_list import PropList
from fabric.actor.core.util.resource_data import ResourceData
from fabric.actor.core.util.resource_type import ResourceType
from fabric.actor.core.util.update_data import UpdateData
from fabric.actor.security.auth_token import AuthToken
from fabric.message_bus.messages.auth_avro import AuthAvro
from fabric.message_bus.messages.delegation_avro import DelegationAvro
from fabric.message_bus.messages.pool_info_avro import PoolInfoAvro
from fabric.message_bus.messages.resource_data_avro import ResourceDataAvro
from fabric.message_bus.messages.resource_set_avro import ResourceSetAvro
from fabric.message_bus.messages.slice_avro import SliceAvro
from fabric.message_bus.messages.term_avro import TermAvro
from fabric.message_bus.messages.update_data_avro import UpdateDataAvro


if TYPE_CHECKING:
    from fabric.actor.core.apis.i_slice import ISlice


class Translate:
    # The direction constants specify the direction of a request. They are used
    # to determine what properties should pass from one actor to another.
    direction_agent = 1
    direction_authority = 2
    direction_return = 3

    @staticmethod
    def translate_udd(*, udd: UpdateData) -> UpdateDataAvro:
        result = UpdateDataAvro()
        result.message = udd.get_message()
        result.failed = udd.is_failed()
        return result

    @staticmethod
    def translate_udd_from_avro(*, udd: UpdateDataAvro) -> UpdateData:
        if udd.failed:
            return UpdateData(message=udd.message)
        else:
            return UpdateData()

    @staticmethod
    def translate_slice(*, slice_id: ID, slice_name: str) -> ISlice:
        if slice_id is None and slice_name is None:
            return None

        if slice_id is None:
            raise ProxyException(Constants.not_specified_prefix.format("Slice id"))

        slice_obj = SliceFactory.create(slice_id=slice_id, name=slice_name)
        return slice_obj

    @staticmethod
    def translate_slice_to_avro(*, slice_obj: ISlice) -> SliceAvro:
        avro_slice = SliceAvro()
        avro_slice.slice_name = slice_obj.get_name()
        avro_slice.guid = str(slice_obj.get_slice_id())
        avro_slice.description = slice_obj.get_description()
        avro_slice.owner = Translate.translate_auth_to_avro(auth=slice_obj.get_owner())

        if slice_obj.get_resource_type() is not None:
            avro_slice.set_resource_type(str(slice_obj.get_resource_type()))

        avro_slice.set_client_slice(slice_obj.is_client())

        if slice_obj.get_graph_id() is not None:
            avro_slice.graph_id = str(slice_obj.get_graph_id())

        return avro_slice

    @staticmethod
    def translate_auth_to_avro(*, auth: AuthToken) -> AuthAvro:
        result = AuthAvro()
        result.name = auth.name
        result.guid = str(auth.guid)
        result.id_token = auth.id_token
        return result

    @staticmethod
    def translate_auth_from_avro(*, auth_avro: AuthAvro) -> AuthToken:
        if auth_avro is None:
            return None
        result = AuthToken()
        result.name = auth_avro.name
        result.guid = auth_avro.guid
        result.id_token = auth_avro.id_token
        return result

    @staticmethod
    def translate_term(*, term: Term) -> TermAvro:
        avro_term = TermAvro()
        if term.get_start_time() is not None:
            avro_term.start_time = ActorClock.to_milliseconds(when=term.get_start_time())
        else:
            term.start_time = 0
        if term.get_end_time() is not None:
            avro_term.end_time = ActorClock.to_milliseconds(when=term.get_end_time())
        else:
            term.end_time = 0
        if term.get_new_start_time() is not None:
            avro_term.new_start_time = ActorClock.to_milliseconds(when=term.get_new_start_time())
        else:
            avro_term.new_start_time = 0
        return avro_term

    @staticmethod
    def translate_resource_data(*, resource_data: ResourceData, direction: int) -> ResourceDataAvro:
        if resource_data is None:
            return None

        avro_rdata = ResourceDataAvro()
        if direction == Translate.direction_agent:
            properties = resource_data.get_request_properties()
            avro_rdata.request_properties = properties

        elif direction == Translate.direction_authority:
            properties = resource_data.get_configuration_properties()
            avro_rdata.config_properties = properties

        elif direction == Translate.direction_return:
            properties = resource_data.get_resource_properties()
            avro_rdata.resource_properties = properties

        else:
            return None
        return avro_rdata

    @staticmethod
    def translate_resource_set(*, resource_set: ResourceSet, direction: int) -> ResourceSetAvro:
        avro_rset = ResourceSetAvro()
        avro_rset.type = str(resource_set.get_type())
        avro_rset.units = resource_set.get_units()
        avro_rset.resource_data = Translate.translate_resource_data(resource_data=resource_set.get_resource_data(),
                                                                    direction=direction)
        return avro_rset

    @staticmethod
    def translate_resource_data_from_avro(*, rdata: ResourceDataAvro) -> ResourceData:
        result = ResourceData()
        properties = rdata.config_properties
        if properties is not None:
            result.configuration_properties = properties

        properties = rdata.request_properties
        if properties is not None:
            result.request_properties = properties

        properties = rdata.resource_properties
        if properties is not None:
            result.resource_properties = properties

        return result

    @staticmethod
    def translate_term_from_avro(*, term: TermAvro) -> Term:
        start_time = None
        end_time = None
        new_start_time = None
        if term.start_time > 0:
            start_time = ActorClock.from_milliseconds(milli_seconds=term.start_time)

        if term.end_time > 0:
            end_time = ActorClock.from_milliseconds(milli_seconds=term.end_time)

        if term.new_start_time > 0:
            new_start_time = ActorClock.from_milliseconds(milli_seconds=term.new_start_time)

        return Term(start=start_time, end=end_time, new_start=new_start_time)

    @staticmethod
    def translate_resource_set_from_avro(*, rset: ResourceSetAvro) -> ResourceSet:
        rdata = Translate.translate_resource_data_from_avro(rdata=rset.resource_data)
        result = ResourceSet(units=rset.units, rtype=ResourceType(resource_type=rset.type), rdata=rdata)
        return result

    @staticmethod
    def attach_properties(*, slice_mng: SliceAvro, slice_obj: ISlice) -> SliceAvro:
        if slice_obj.get_request_properties() is not None:
            slice_mng.set_request_properties(slice_obj.get_request_properties())

        if slice_obj.get_resource_properties() is not None:
            slice_mng.set_resource_properties(slice_obj.get_resource_properties())

        if slice_obj.get_config_properties() is not None:
            slice_mng.set_config_properties(slice_obj.get_config_properties())

        if slice_obj.get_local_properties() is not None:
            slice_mng.set_config_properties(slice_obj.get_local_properties())

        return slice_mng

    @staticmethod
    def fill_slices(*, slice_list: list, full: bool) -> list:
        result = []
        for s in slice_list:
            slice_obj = SliceFactory.create_instance(properties=s)
            if slice_obj is not None:
                ss = Translate.translate_slice_to_avro(slice_obj=slice_obj)
                if full:
                    ss = Translate.attach_properties(slice_mng=ss, slice_obj=slice_obj)
                result.append(ss)
        return result

    @staticmethod
    def absorb_properties(*, slice_mng: SliceAvro, slice_obj: ISlice) -> ISlice:
        slice_obj.set_local_properties(value=PropList.merge_properties(
            incoming=slice_mng.get_local_properties(), outgoing=slice_obj.get_local_properties()))

        slice_obj.set_config_properties(value=PropList.merge_properties(
            incoming=slice_mng.get_config_properties(), outgoing=slice_obj.get_config_properties()))

        slice_obj.set_request_properties(value=PropList.merge_properties(
            incoming=slice_mng.get_request_properties(), outgoing=slice_obj.get_request_properties()))

        slice_obj.set_resource_properties(value=PropList.merge_properties(
            incoming=slice_mng.get_resource_properties(), outgoing=slice_obj.get_resource_properties()))

        return slice_obj

    @staticmethod
    def translate_delegation_to_avro(*, delegation: IDelegation) -> DelegationAvro:
        avro_delegation = DelegationAvro()
        avro_delegation.delegation_id = delegation.get_delegation_id()
        avro_delegation.slice = Translate.translate_slice_to_avro(slice_obj=delegation.get_slice_object())
        if delegation.get_graph() is not None:
            avro_delegation.graph = delegation.get_graph().serialize_graph()
        return avro_delegation

    @staticmethod
    def translate_to_pool_info(*, query_response: dict) -> PoolInfoAvro:
        bqm = query_response.get(Constants.broker_query_model, None)
        pool_info = PoolInfoAvro()
        pool_info.type = Constants.pool_type
        pool_info.name = Constants.broker_query_model
        if bqm is not None:
            pool_info.properties = {Constants.broker_query_model: bqm}
        return pool_info



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING

from fabric.actor.core.manage.error import Error
from fabric.actor.core.apis.i_component import IComponent
from fabric.message_bus.messages.result_avro import ResultAvro

if TYPE_CHECKING:
    from fabric.actor.core.manage.management_object import ManagementObject
    from fabric.actor.security.auth_token import AuthToken


class LocalProxy(IComponent):
    def __init__(self, *, manager: ManagementObject, auth: AuthToken):
        self.manager = manager
        self.auth = auth
        self.last_status = None
        self.last_exception = None

    def clear_last(self):
        self.last_exception = None
        self.last_status = ResultAvro()

    def get_last_error(self) -> Error:
        return Error(status=self.last_status, e=self.last_exception)

    def get_manager_object(self) -> ManagementObject:
        return self.manager

    def get_protocols(self) -> list:
        return self.manager.get_proxies()

    def get_type_id(self) -> str:
        return str(self.manager.get_type_id())

    def get_first(self, *, result_list: list):
        if result_list is not None and len(result_list) > 0:
            return result_list.__iter__().__next__()
        return None


#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

from fabric.actor.core.common.exceptions import PolicyException

if TYPE_CHECKING:
    from fabric.actor.core.util.resource_type import ResourceType
    from fabric.actor.core.apis.i_client_reservation import IClientReservation
    from fabric.actor.core.common.resource_pool_descriptor import ResourcePoolDescriptor


class InventoryForType:
    def __init__(self):
        self.resource_type = None
        self.properties = {}
        self.source = None
        self.resource_pool_descriptor = None

    def get_type(self) -> ResourceType:
        return self.resource_type

    def set_type(self, *, rtype: ResourceType):
        self.resource_type = rtype

    def donate(self, *, source: IClientReservation):
        if self.source is not None:
            raise PolicyException("This inventory pool already has a source.")
        self.source = source

    def get_source(self) -> IClientReservation:
        return self.source

    def set_descriptor(self, *, rpd: ResourcePoolDescriptor):
        self.resource_pool_descriptor = rpd

    def get_descriptor(self) -> ResourcePoolDescriptor:
        return self.resource_pool_descriptor

    def get_properties(self) -> dict:
        return self.properties

    @abstractmethod
    def allocate(self, *, count: int, request: dict, resource: dict = None) -> dict:
        """
        Allocates the specified number of units given the client request
        properties. This method is called for new ticketing/extending reservations.
        @param count how many units to allocate
        @param request request properties
        @param resource what is currently allocated
        @return the resource properties to be passed back to the client
        """

    @abstractmethod
    def allocate_revisit(self, *, count: int, resource: dict):
        """
        Called during revisit to indicate that a ticketed reservation is being
        recovered.
        @param count number of units
        @param resource resource properties
        """

    @abstractmethod
    def free(self, *, count: int, request: dict = None, resource: dict = None) -> dict:
        """
        Frees the specified number of resource units.
        @param count number of units
        @param request request properties
        @param resource resource properties
        @return new resource properties
        """

    @abstractmethod
    def get_free(self) -> int:
        """
        Returns the number of free units in the inventory pool.
        @return number of free units
        """

    @abstractmethod
    def get_allocated(self) -> int:
        """
        Returns the number of allocated units from this invento
        @return number of allocated units
        """



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING, List

from fabric.actor.core.apis.i_actor_runnable import IActorRunnable
from fabric.actor.core.common.constants import Constants, ErrorCodes
from fabric.actor.core.common.exceptions import ReservationNotFoundException, SliceNotFoundException, \
    DelegationNotFoundException, ManageException
from fabric.actor.core.delegation.delegation_factory import DelegationFactory
from fabric.actor.core.kernel.reservation_factory import ReservationFactory
from fabric.actor.core.kernel.reservation_states import ReservationStates, ReservationPendingStates
from fabric.actor.core.kernel.slice_factory import SliceFactory
from fabric.actor.core.manage.converter import Converter
from fabric.actor.core.manage.management_object import ManagementObject
from fabric.actor.core.manage.management_utils import ManagementUtils
from fabric.actor.core.manage.proxy_protocol_descriptor import ProxyProtocolDescriptor
from fabric.actor.core.apis.i_actor_management_object import IActorManagementObject
from fabric.actor.security.acess_checker import AccessChecker
from fabric.actor.security.pdp_auth import ActionId, ResourceType
from fabric.message_bus.messages.reservation_mng import ReservationMng
from fabric.message_bus.messages.reservation_state_avro import ReservationStateAvro
from fabric.actor.core.manage.messages.result_event_mng import ResultEventMng
from fabric.message_bus.messages.result_delegation_avro import ResultDelegationAvro
from fabric.message_bus.messages.result_reservation_avro import ResultReservationAvro
from fabric.message_bus.messages.result_reservation_state_avro import ResultReservationStateAvro
from fabric.message_bus.messages.result_string_avro import ResultStringAvro
from fabric.actor.core.proxies.kafka.translate import Translate
from fabric.actor.core.registry.actor_registry import ActorRegistrySingleton
from fabric.actor.core.util.all_actor_events_filter import AllActorEventsFilter
from fabric.actor.core.util.id import ID
from fabric.actor.security.auth_token import AuthToken
from fabric.message_bus.messages.result_avro import ResultAvro
from fabric.message_bus.messages.result_slice_avro import ResultSliceAvro
from fabric.message_bus.messages.slice_avro import SliceAvro

if TYPE_CHECKING:
    from fabric.actor.core.apis.i_actor import IActor
    from fabric.actor.core.apis.i_slice import ISlice


class ActorManagementObject(ManagementObject, IActorManagementObject):
    def __init__(self, *, actor: IActor = None):
        super().__init__()
        self.actor = actor
        self.db = None
        if actor is not None:
            self.set_actor(actor=actor)

    def register_protocols(self):
        from fabric.actor.core.manage.local.local_actor import LocalActor
        local = ProxyProtocolDescriptor(protocol=Constants.protocol_local, proxy_class=LocalActor.__name__,
                                        proxy_module=LocalActor.__module__)

        from fabric.actor.core.manage.kafka.kafka_actor import KafkaActor
        kafka = ProxyProtocolDescriptor(protocol=Constants.protocol_kafka, proxy_class=KafkaActor.__name__,
                                        proxy_module=KafkaActor.__module__)

        self.proxies = []
        self.proxies.append(local)
        self.proxies.append(kafka)

    def save(self) -> dict:
        properties = super().save()
        properties[Constants.property_class_name] = ActorManagementObject.__name__
        properties[Constants.property_module_name] = ActorManagementObject.__module__

        return properties

    def recover(self):
        actor_name = None
        if Constants.property_actor_name in self.serial:
            actor_name = self.serial[Constants.property_actor_name]
        else:
            raise ManageException(Constants.not_specified_prefix.format("actor name"))

        actor = ActorRegistrySingleton.get().get_actor(actor_name_or_guid=actor_name)

        if actor is None:
            raise ManageException(Constants.object_not_found.format("Managed Object", actor_name))

        self.set_actor(actor=actor)

    def set_actor(self, *, actor: IActor):
        if self.actor is None:
            self.actor = actor
            self.db = actor.get_plugin().get_database()
            self.logger = actor.get_logger()
            self.id = actor.get_guid()

    def get_slices(self, *, slice_id: ID, caller: AuthToken, id_token: str = None) -> ResultSliceAvro:
        result = ResultSliceAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
        else:
            slice_list = None
            try:
                if id_token is not None:
                    AccessChecker.check_access(action_id=ActionId.query, resource_type=ResourceType.slice,
                                               token=id_token, logger=self.logger, actor_type=self.actor.get_type(),
                                               resource_id=str(slice_id))
                try:
                    slice_list = self.db.get_slices(slice_id=slice_id)
                except Exception as e:
                    self.logger.error("getSlices:db access {}".format(e))
                    result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                    result.status.set_message(ErrorCodes.ErrorDatabaseError.name)
                    result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                if slice_list is not None:
                    result.slices = Translate.fill_slices(slice_list=slice_list, full=True)
                else:
                    result.status.set_code(ErrorCodes.ErrorNoSuchSlice.value)
                    result.status.set_message(ErrorCodes.ErrorNoSuchSlice.name)
            except Exception as e:
                self.logger.error("getSlices {}".format(e))
                result.status.set_code(ErrorCodes.ErrorInternalError.value)
                result.status.set_message(ErrorCodes.ErrorInternalError.name)
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
        return result

    def add_slice(self, *, slice_mng: SliceAvro, caller: AuthToken) -> ResultStringAvro:
        result = ResultStringAvro()
        result.status = ResultAvro()

        if slice_mng is None or caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)

        else:
            try:
                data = Converter.get_resource_data(slice_mng=slice_mng)
                slice_obj = SliceFactory.create(slice_id=ID(), name=slice_mng.get_slice_name(), data=data)

                slice_obj.set_description(description=slice_mng.get_description())
                slice_obj.set_owner(owner=self.actor.get_identity())
                slice_obj.set_inventory(value=True)

                class Runner(IActorRunnable):
                    def __init__(self, *, actor: IActor):
                        self.actor = actor

                    def run(self):
                        try:
                            self.actor.register_slice(slice_object=slice_obj)
                        except Exception as e:
                            self.actor.get_plugin().release_slice(slice_obj=slice_obj)
                            raise e

                        return None

                self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))

                result.set_result(str(slice_obj.get_slice_id()))
            except Exception as e:
                self.logger.error("addslice: {}".format(e))
                result.status.set_code(ErrorCodes.ErrorInternalError.value)
                result.status.set_message(ErrorCodes.ErrorInternalError.name)
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def remove_slice(self, *, slice_id: ID, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()
        if slice_id is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    self.actor.remove_slice_by_slice_id(slice_id=slice_id)
                    return None

            self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))
        except Exception as e:
            self.logger.error("remove_slice: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)
        return result

    def update_slice(self, *, slice_mng: SliceAvro, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()
        if slice_mng is None or slice_mng.get_slice_id() is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            slice_id = ID(uid=slice_mng.get_slice_id())

            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    result = ResultAvro()
                    slice_obj = self.actor.get_slice(slice_id=slice_id)
                    if slice_obj is None:
                        result.set_code(ErrorCodes.ErrorNoSuchSlice.value)
                        result.set_message(ErrorCodes.ErrorNoSuchSlice.name)
                        return result
                    slice_obj = ManagementUtils.update_slice(slice_obj=slice_obj, slice_mng=slice_mng)

                    try:
                        self.actor.get_plugin().get_database().update_slice(slice_object=slice_obj)
                    except Exception as e:
                        print("Could not commit slice update {}".format(e))
                        result.set_code(ErrorCodes.ErrorDatabaseError.value)
                        result.set_message(ErrorCodes.ErrorDatabaseError.name)
                        result = ManagementObject.set_exception_details(result=result, e=e)

                    return result

            result = self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))
        except Exception as e:
            self.logger.error("update_slice: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def _get_slice_by_id(self, *, id: int) -> ISlice:
        ss = self.db.get_slice_by_id(id=id)
        slice_obj = SliceFactory.create_instance(properties=ss)
        return slice_obj

    def get_slice_by_guid(self, *, guid: str, id_token: str = None) -> ISlice:
        ss = self.db.get_slice(slice_guid=guid)
        slice_obj = SliceFactory.create_instance(properties=ss)
        return slice_obj

    def get_reservations(self, *, caller: AuthToken, id_token: str = None, state: int = None,
                         slice_id: ID = None, rid: ID = None) -> ResultReservationAvro:
        result = ResultReservationAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            AccessChecker.check_access(action_id=ActionId.query, resource_type=ResourceType.sliver,
                                       token=id_token, logger=self.logger, actor_type=self.actor.get_type(),
                                       resource_id=str(rid))
            res_list = None
            try:
                if rid is not None:
                    res = self.db.get_reservation(rid)
                    if res is not None:
                        res_list = [res]
                    else:
                        result.status.set_code(ErrorCodes.ErrorNoSuchReservation.value)
                        result.status.set_message(ErrorCodes.ErrorNoSuchReservation.name)
                elif slice_id is not None and state is not None:
                    res_list = self.db.get_reservations_by_slice_id_state(slc_guid=slice_id, rsv_state=state)
                elif slice_id is not None:
                    res_list = self.db.get_reservations_by_slice_id(slc_guid=slice_id)
                elif state is not None:
                    res_list = self.db.get_reservations_by_state(rsv_state=state)
                else:
                    res_list = self.db.get_reservations()
            except Exception as e:
                self.logger.error("getReservations:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.name)
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)

            if res_list is not None:
                result.reservations = []
                for r in res_list:
                    slice_id = ReservationFactory.get_slice_id(properties=r)
                    if slice_id is None:
                        self.logger.error("Inconsistent state reservation does not belong to a slice: {}".format(r))

                    slice_obj = None
                    if slice_id is not None:
                        slice_obj = self._get_slice_by_id(id=slice_id)
                    rsv_obj = ReservationFactory.create_instance(properties=r, actor=self.actor, slice_obj=slice_obj,
                                                                 logger=self.actor.get_logger())
                    if rsv_obj is not None:
                        rr = Converter.fill_reservation(reservation=rsv_obj, full=False)
                        result.reservations.append(rr)
        except ReservationNotFoundException as e:
            self.logger.error("getReservations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorNoSuchReservation.value)
            result.status.set_message(e.text)
        except Exception as e:
            self.logger.error("getReservations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def remove_reservation(self, *, caller: AuthToken, rid: ID) -> ResultAvro:
        result = ResultAvro()

        if rid is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    self.actor.remove_reservation(rid=rid)
                    return None

            self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))
        except ReservationNotFoundException as e:
            self.logger.error("remove_reservation: {}".format(e))
            result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
            result.set_message(e.text)
        except Exception as e:
            self.logger.error("remove_reservation: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def close_reservation(self, *, caller: AuthToken, rid: ID) -> ResultAvro:
        result = ResultAvro()

        if rid is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    self.actor.close_by_rid(rid=rid)
                    return True

            self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))
        except ReservationNotFoundException as e:
            self.logger.error("close_reservation: {}".format(e))
            result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
            result.set_message(e.text)
        except Exception as e:
            self.logger.error("close_reservation: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def close_slice_reservations(self, *, caller: AuthToken, slice_id: ID) -> ResultAvro:
        result = ResultAvro()

        if slice_id is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    self.actor.close_slice_reservations(slice_id=slice_id)
                    return None

            self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))
        except SliceNotFoundException as e:
            self.logger.error("close_slice_reservations: {}".format(e))
            result.set_code(ErrorCodes.ErrorNoSuchSlice.value)
            result.set_message(e.text)
            result = ManagementObject.set_exception_details(result=result, e=e)
        except Exception as e:
            self.logger.error("close_slice_reservations: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)
        return result

    def get_actor(self) -> IActor:
        return self.actor

    def get_actor_name(self) -> str:
        if self.actor is not None:
            return self.actor.get_name()

        return None

    def update_reservation(self, *, reservation: ReservationMng, caller: AuthToken) -> ResultAvro:
        result = ResultAvro()
        if reservation is None or caller is None:
            result.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            rid = ID(uid=reservation.get_reservation_id())

            class Runner(IActorRunnable):
                def __init__(self, *, actor: IActor):
                    self.actor = actor

                def run(self):
                    result = ResultAvro()
                    r = self.actor.get_reservation(rid=rid)
                    if r is None:
                        result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
                        result.set_message(ErrorCodes.ErrorNoSuchReservation.name)
                        return result
                    r = ManagementUtils.update_reservation(res_obj=r, rsv_mng=reservation)
                    r.set_dirty()

                    try:
                        self.actor.get_plugin().get_database().update_reservation(reservation=r)
                    except Exception as e:
                        print("Could not commit slice update {}".format(e))
                        result.set_code(ErrorCodes.ErrorDatabaseError.value)
                        result.set_message(ErrorCodes.ErrorDatabaseError.name)
                        result = ManagementObject.set_exception_details(result=result, e=e)
                    return result

            result = self.actor.execute_on_actor_thread_and_wait(runnable=Runner(actor=self.actor))
        except ReservationNotFoundException as e:
            self.logger.error("update_reservation: {}".format(e))
            result.set_code(ErrorCodes.ErrorNoSuchReservation.value)
            result.set_message(e.text)
        except Exception as e:
            self.logger.error("update_reservation: {}".format(e))
            result.set_code(ErrorCodes.ErrorInternalError.value)
            result.set_message(ErrorCodes.ErrorInternalError.name)
            result = ManagementObject.set_exception_details(result=result, e=e)

        return result

    def get_reservation_state_for_reservations(self, *, caller: AuthToken, rids: List[str],
                                               id_token: str = None) -> ResultReservationStateAvro:
        result = ResultReservationStateAvro()
        result.status = ResultAvro()

        if rids is None or len(rids) == 0 or caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        for rid in rids:
            if rid is None:
                result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
                result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
                return result

        try:
            AccessChecker.check_access(action_id=ActionId.query, resource_type=ResourceType.sliver,
                                       token=id_token, logger=self.logger, actor_type=self.actor.get_type())
            res_list = None
            try:
                res_list = self.db.get_reservations_by_rids(rid=rids)
            except Exception as e:
                self.logger.error("get_reservation_state_for_reservations:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.name)
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)
                return result

            if len(res_list) == len(rids):
                result.reservation_states = Converter.fill_reservation_state(res=res_list)
            elif len(res_list) > len(rids):
                raise ManageException("The database provided too many records")
            else:
                i = 0
                j = 0
                while i < len(rids):
                    properties = res_list[j]
                    if rids[i] == ReservationFactory.get_reservation_id(properties=properties):
                        result.reservation_states.append(Converter.fill_reservation_state(res=res_list))
                        j += 1

                    else:
                        state = ReservationStateAvro()
                        state.set_state(ReservationStates.Unknown.value)
                        state.set_pending_state(ReservationPendingStates.Unknown.value)
                        result.reservation_states.append(state)
                    i += 1
        except ReservationNotFoundException as e:
            self.logger.error("get_reservation_state_for_reservations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorNoSuchReservation.value)
            result.status.set_message(e.text)
        except Exception as e:
            self.logger.error("get_reservation_state_for_reservations {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result

    def get_delegations(self, *, caller: AuthToken, id_token: str = None, slice_id: ID = None,
                        did: ID = None) -> ResultDelegationAvro:
        result = ResultDelegationAvro()
        result.status = ResultAvro()

        if caller is None:
            result.status.set_code(ErrorCodes.ErrorInvalidArguments.value)
            result.status.set_message(ErrorCodes.ErrorInvalidArguments.name)
            return result

        try:
            if id_token is not None:
                AccessChecker.check_access(action_id=ActionId.query, resource_type=ResourceType.delegation,
                                           token=id_token, logger=self.logger, actor_type=self.actor.get_type())
            res_list = None
            try:
                if did is not None:
                    res = self.db.get_delegation(dlg_graph_id=did)
                    if res is not None:
                        res_list = [res]
                    else:
                        result.status.set_code(ErrorCodes.ErrorNoSuchDelegation.value)
                        result.status.set_message(ErrorCodes.ErrorNoSuchDelegation.name)
                elif slice_id is not None:
                    res_list = self.db.get_delegations_by_slice_id(slice_id=slice_id)
                else:
                    res_list = self.db.get_delegations()
            except Exception as e:
                self.logger.error("getDelegations:db access {}".format(e))
                result.status.set_code(ErrorCodes.ErrorDatabaseError.value)
                result.status.set_message(ErrorCodes.ErrorDatabaseError.name)
                result.status = ManagementObject.set_exception_details(result=result.status, e=e)

            if res_list is not None:
                result.delegations = []
                for r in res_list:
                    slice_id = ReservationFactory.get_delegation_slice_id(properties=r)
                    if slice_id is None:
                        self.logger.error("Inconsistent state delegation does not belong to a slice: {}".format(r))

                    slice_obj = None
                    if slice_id is not None:
                        slice_obj = self._get_slice_by_id(id=slice_id)
                    dlg_obj = DelegationFactory.create_instance(properties=r, actor=self.actor, slice_obj=slice_obj,
                                                                logger=self.actor.get_logger())
                    if dlg_obj is not None:
                        rr = Translate.translate_delegation_to_avro(delegation=dlg_obj)
                        result.delegations.append(rr)
        except DelegationNotFoundException as e:
            self.logger.error("getDelegations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorNoSuchDelegation.value)
            result.status.set_message(e.text)
        except Exception as e:
            self.logger.error("getDelegations: {}".format(e))
            result.status.set_code(ErrorCodes.ErrorInternalError.value)
            result.status.set_message(ErrorCodes.ErrorInternalError.name)
            result.status = ManagementObject.set_exception_details(result=result.status, e=e)

        return result



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from enum import Enum


class ResourcePoolAttributeType(Enum):
    """
    Attribute Type Enumeration
    """
    UNDEFINED = 1
    INTEGER = 2
    FLOAT = 3
    STRING = 4
    NEO4J = 5
    CLASS = 6


class ResourcePoolAttributeDescriptor:
    """
    Pool Attribute Description
    """
    property_type = "type"
    property_unit = "unit"
    property_value = "value"
    property_label = "label"

    def __init__(self):
        self.key = None
        self.type = None
        self.value = None
        self.label = None
        self.unit = None

    def get_type(self) -> ResourcePoolAttributeType:
        """
        Get type
        @return type
        """
        return self.type

    def set_type(self, *, rtype: ResourcePoolAttributeType):
        """
        Set type
        @param rtype rtype
        """
        self.type = rtype

    def get_value(self) -> str:
        """
        Get value
        @return value
        """
        return self.value

    def get_int_value(self) -> int:
        """
        Get Integer value
        @return in value
        """
        return int(self.value)

    def set_value(self, *, value: str):
        """
        Set value
        @param value value
        """
        self.value = value

    def get_key(self) -> str:
        """
        Return key
        @return key
        """
        return self.key

    def set_key(self, *, value: str):
        """
        Set key
        @param value value
        """
        self.key = value

    def set_label(self, *, label: str):
        """
        Set label
        @param label label
        """
        self.label = label

    def get_label(self) -> str:
        """
        Get label
        @return label
        """
        return self.label

    def get_unit(self) -> str:
        """
        Get unit
        @return unit
        """
        return self.unit

    def set_unit(self, *, unit: str):
        """
        Set unit
        @param unit unit
        """
        self.unit = unit

    def __str__(self):
        return "{}:{}:{}".format(self.key, self.value, self.type)

    def save(self, *, properties: dict, prefix: str) -> dict:
        """
        Save in properties
        @param properties properties
        @param prefix prefix
        @return saved properties
        """
        if prefix is None:
            prefix = ""

        properties[prefix + self.property_type] = str(self.type.value)
        if self.value is not None:
            properties[prefix + self.property_value] = self.value

        if self.label is not None:
            properties[prefix + self.property_label] = self.label

        if self.unit is not None:
            properties[prefix + self.property_unit] = self.unit

        return properties

    def reset(self, *, properties: dict, prefix: str):
        """
        Reset the properties
        @param properties properties
        @param prefix prefix
        """
        if prefix is None:
            prefix = ""

        self.type = ResourcePoolAttributeType(int(properties[prefix + self.property_type]))
        self.value = properties[prefix + self.property_value]
        if prefix + self.property_label in properties:
            self.label = properties[prefix + self.property_label]
        if prefix + self.property_unit in properties:
            self.unit = properties[prefix + self.property_unit]



# coding: utf-8

from setuptools import setup, find_packages

from fabric import actor

NAME = "fabric-cf"
VERSION = "0.5"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.read()

setup(
    name=NAME,
    version=VERSION,
    description="Fabric Control Framework",
    author="Komal Thareja",
    author_email="kthare10@renci.org",
    url="https://github.com/fabric-testbed/ControlFramework",
    keywords=["Swagger", "Fabric Control Framework"],
    install_requires=requirements,
    setup_requires=requirements,
    packages=find_packages(),
    include_package_data=True,
    data_files=[(actor.CONF_DIR, [actor.CONF_FILE]), (actor.LOG_DIR, [])],
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7'
)



#!/usr/bin/env python3
# MIT License
#
# Copyright (c) 2020 FABRIC Testbed
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#
# Author: Komal Thareja (kthare10@renci.org)
from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime
from fabric.actor.core.apis.i_reservation import IReservation, ReservationCategory
from fabric.actor.core.common.constants import Constants
from fabric.actor.core.common.exceptions import ResourcesException
from fabric.actor.core.util.resource_data import ResourceData

if TYPE_CHECKING:
    from fabric.actor.core.time.term import Term
    from fabric.actor.core.util.id import ID
    from fabric.actor.core.util.notice import Notice
    from fabric.actor.core.apis.i_concrete_set import IConcreteSet
    from fabric.actor.core.util import resource_type
    from fabric.actor.core.core.ticket import Ticket


class ResourceSet:
    """
    ResourceSet is an abstract set of resources describing some number of
    resource units of a given type, e.g., to represent a resource request. A
    ResourceSet with an attached ConcreteSet is "concrete." The ConcreteSet binds
    real resources (or a promise) for some or all of the abstract units in the
    set: for example, a concrete ResourceSet can represent a ticket or
    lease. Adding or removing concrete resources does not affect the number of
    abstract units. If there are fewer concrete units than abstract units, the
    set has a "deficit".

    An "elastic" ResourceSet may be filled at less than its full request, and it
    may change size on extends. An actor may modify an elastic ResourceSet on an
    active ReservationClient by calling "flex", if there is no pending operation
    in progress, e.g., in preparation for an elastic extend. This class updates
    the unit count to match the concrete resources on each reserve or extend (on
    a server), or update (on a client).

    Operations on the ConcreteSet through this class may drive probes and state
    transitions on the underlying resources transferred in and out of the
    ResourceSet (e.g., node configuration and node reboot for a COD authority, or
    resource membership changes on a orchestrator). ConcreteSets are
    responsible for their own synchronization: calls to ConcreteSet go through
    pre-op "prepare" or post-op "service" methods in this class, which may block
    and should not hold any higher-level locks. Most other operations are called
    through Mapper or the Reservation class with the Manager lock held.

    Implementation notes
    The unit count is updated immediately to reflect additions or deletions
    from the set. Updates to the unit count must occur only in the locked
    methods. Configuration actions on the ConcreteSet (e.g., as resources join
    and leave the set) must occur only in unlocked methods (e.g., "service"). A
    tricky part is flex(), which updates abstract count to reflect a new request:
    it is unlocked, which could race with an incoming unsolicited lease (which
    are currently allowed), or with overlapping requests on the same set (which
    are currently not allowed).
    ResourceSet was conceived as supporting methods that are independent of
    context and type of ConcreteSet. That ideal has eroded somewhat, and some key
    fields and methods are specific to a particular context or role. Someday it
    may be useful to break this into subclasses.
    Currently leases are validated only with validateIncoming(). There may be
    some additional checks to enforce.
    No changes to ResourceData on merges. Needs thought and documentation. We
    should remove the properties argument on ConcreteSet.change.
    The 'null ticket corner case' (see above) is a source of complexity, and
    should be cleaned up.
    Calls that "reach around" ResourceSet to the concrete set are
    discouraged/deprecated.
    """
    def __init__(self, *, concrete: IConcreteSet = None, gained: IConcreteSet = None,
                 lost: IConcreteSet = None, modified: IConcreteSet = None,
                 rtype: resource_type = None, rdata: ResourceData = None, units: int = None):
        # What type of resources does this set contain. The meaning/assignment of
        # type values is an externally defined convention of interacting actors.
        self.type = rtype
        # How many units (abstract) the set contains. This count reflects the
        # resources intended or requested for this set. For an active reservation
        # in steady state, the unit count will typically match the number of
        # concrete resources, but it might not match if the resource set is in flux
        # for some reason. For an inventory set, the abstract count reflects the
        # original size of the inventory, independent of any allocations extracted
        # from it.
        self.units = 0
        # Concrete resources.
        self.resources = concrete
        if concrete is not None:
            self.units = concrete.get_units()
        else:
            self.units = units
        # ResourceData is a property list for this ResourceSet. E.g., request
        # attributes, node attributes, etc.
        self.properties = rdata
        # The previous value of the properties list. This is essential for
        # supporting recovery on Authority.
        self.previous_properties = None
        # A set of resources recently ejected from the resource set, pending
        # processing, e.g., by a "probe" method.
        self.released = None
        # A recent update to the concrete resource set, pending processing by a
        # "service" method. Client-side only (i.e., for ticket or lease updates
        # through callback interface).
        self.updated = None
        # Recent additions of resources to the concrete set, pending processing by
        # a "service" method. Only for authority role only.
        self.gained = gained
        # Recently lost resources pending processing by a "service" method. Only
        # for authority role only.
        self.lost = lost
        # Recently changed resources pending a processing by a "service" method.
        # set.
        self.modified = modified
        # Reservation with which this set is associated.
        self.rid = None
        self.is_closing = False

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['released']
        del state['updated']
        del state['gained']
        del state['lost']
        del state['modified']
        del state['rid']
        del state['is_closing']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.released = None
        self.updated = None
        self.gained = None
        self.lost = None
        self.modified = None
        self.rid = None
        self.is_closing = False

    def restore(self, *, rid: ID):
        """
        Restore post stateful restart
        @param rid reservation id
        """
        self.rid = rid

    def abstract_clone(self):
        """
        Clones the set, but without any of the concrete sets. Used on Authority
        and Service Manager to create a ResourceSet to hold resources, given a
        ResourceSet holding a ticket. All properties are preserved (cloned).
        @return a resources set that is a copy of the current but without any concrete sets.
        """
        rd = self.properties.clone()
        result = ResourceSet(units=self.units, rtype=self.type, rdata=rd)
        return result

    def collect_released(self) -> ResourceSet:
        """
        Any units that fail or are rejected/released asynchronously accumulate
        within the ConcreteSet until collected. These are cached by a
        prepareProbe.
        @returns a ResourceSet
        @raises Exception in case of error
        """
        rset = None
        if self.released is not None:
            rset = ResourceSet(concrete=self.released, rtype=self.type, rdata=ResourceData())
        return rset

    def merge_properties(self, *, reservation: IReservation, resource_set):
        """
        Merge properties
        @param reservation reservation
        @param resource_set resource set
        """
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.invalid_argument)

        if self.properties is None:
            self.properties = ResourceData()

        if reservation.get_category() == ReservationCategory.Client:
            # On a orchestrator we only take resource properties from
            # the broker when we receive the initial ticket update. The
            # current assumption is that resource properties are constant
            # over the lifetime of a reservation and are supplied by the
            # broker.
            from fabric.actor.core.core.ticket import Ticket
            if self.resources is None and isinstance(resource_set.get_resources(), Ticket):
                self.properties.resource_properties = ResourceData.merge_properties(
                    from_props=resource_set.get_resource_properties(),
                    to_props=self.properties.resource_properties)

        elif reservation.get_category() == ReservationCategory.Broker:
            # On a broker we will store in the resulting resource set only
            # the resource properties, so that when we generate the update
            # those can be send back to the client. Since the assumption is
            # that resource properties are constant over the lifetime of a
            # reservation, we will do this only for the initial ticket
            # allocation. Ticket extentions will not result in the addition
            # of new properties.
            if self.resources is None:
                self.properties.resource_properties = ResourceData.merge_properties(
                    from_props=resource_set.get_resource_properties(),
                    to_props=self.properties.resource_properties)

        elif reservation.get_category() == ReservationCategory.Authority:
            # On an authority we will take the local (provided by the
            # authority) and configuration, provided by the client,
            # properties. Local and configuration properties need to be
            # merged for each update, as they may contain additional
            # information required by the configuration handlers.
            self.properties.local_properties = ResourceData.merge_properties(
                from_props=resource_set.get_local_properties(),
                to_props=self.properties.local_properties)

            self.properties.configuration_properties = ResourceData.merge_properties(
                from_props=resource_set.get_config_properties(),
                to_props=self.properties.configuration_properties)

    def delta_update(self, *, reservation: IReservation, resource_set):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.invalid_argument)

        if self.resources is None:
            # in case of close for a canceled reservation.
            if resource_set.gained is None:
                return
            # first time we give concrete resources to this resource set
            # Since this set has no concrete resources, we can only gain
            # resources. Lost and modified have no meaning in this case. Assert just in case
            if resource_set.lost is not None or resource_set.modified is not None:
                return
            # take the units and type
            self.units = resource_set.gained.get_units()
            self.type = resource_set.type
            self.resources = resource_set.gained.clone_empty()
            self.resources.setup(reservation=reservation)
            self.merge_properties(reservation=reservation, resource_set=resource_set)
            self.gained = resource_set.gained
        else:
            self.type = resource_set.type
            difference = 0
            if resource_set.gained is None or resource_set.lost is not None or resource_set.modified is not None:
                raise ResourcesException("Internal Error: service overrun in hardChange")

            if resource_set.gained is not None:
                self.gained = resource_set.gained
                difference = resource_set.gained.get_units()

            if resource_set.lost is not None:
                self.lost = resource_set.lost
                difference -= resource_set.lost.get_units()

            if resource_set.modified is not None:
                self.modified = resource_set.modified

            self.units += difference
            self.previous_properties = self.properties
            self.merge_properties(reservation=reservation, resource_set=resource_set)

    def fix_abstract_units(self):
        """
        Sets the number of abstract units to equal the number of concrete units.
        """
        if self.resources is None:
            self.units = self.resources.get_units()
        else:
            self.units = 0

    def full_update(self, *, reservation: IReservation, resource_set):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.invalid_argument)

        # take the units and the type
        self.units = resource_set.units
        self.type = resource_set.type
        # take in the properties
        self.previous_properties = self.properties
        self.merge_properties(reservation=reservation, resource_set=resource_set)

        # make a concrete set if the current concrete set is None
        if self.resources is None:
            self.resources = resource_set.resources.clone_empty()
            self.resources.setup(reservation=reservation)
        # remember the update so that it can be processed later
        self.updated = resource_set.resources

    def get_concrete_units(self, *, when: datetime = None) -> int:
        """
        Estimate the concrete resource units the resource set will contain at the
        specified date.
        @params when: the date
        @returns number of concrete units
        """
        if self.resources is None:
            return 0
        elif when is None:
            return self.resources.get_units()
        else:
            return self.resources.holding(when=when)

    def get_config_properties(self) -> dict:
        """
        Returns the configuration properties list.
        @returns configuration properties list. Can be null.
        """
        if self.properties is not None:
            return self.properties.get_configuration_properties()
        return None

    def get_deficit(self) -> int:
        """
        Returns the number of concrete units needed or in excess in this resource
        set.
        @returns number of units in excess or needed
        """
        result = self.units
        if self.resources is not None:
            result -= self.resources.get_units()

        return result

    def get_local_properties(self) -> dict:
        """
        Returns the local properties list
        @returns local properties list. Can be null.
        """
        if self.properties is not None:
            return self.properties.get_local_properties()
        return None

    def get_notices(self) -> Notice:
        """
        Returns a string of notices or events pertaining to the underlying
        resources. The event notices are consumed: subsequent calls return only
        new information. May return null.
        @returns Notice
        """
        if self.resources is None:
            return None
        return self.resources.get_notices()

    def get_request_properties(self):
        """
        Returns the request properties.
        @returns request properties list. Can be null.
        """
        if self.properties is not None:
            self.properties.get_request_properties()
        return None

    def get_reservation_id(self) -> ID:
        """
        Returns the reservation identifier attached to this resource set.
        @returns reservation identifier
        """
        return self.rid

    def get_resource_data(self) -> ResourceData:
        """
        Returns the properties of this resource set.
        @returns set properties
        """
        return self.properties

    def get_resource_properties(self):
        """
        Returns the resource properties.
        @returns resource properties list. Can be null.
        """
        if self.properties is not None:
            return self.properties.get_resource_properties()
        return None

    def get_resources(self) -> IConcreteSet:
        """
        Returns the concrete resources.
        @returns concrete resource set
        """
        return self.resources

    def get_site_proxy(self):
        """
        Returns a proxy to the site authority, which owns the resources
        represented in the set.
        @returns site authority proxy.
        @raises Exception in case of error
        """
        if self.resources is not None:
            return self.resources.get_site_proxy()
        return None

    def get_type(self) -> resource_type:
        """
        Returns the resource type of the set.
        @returns resource type
        """
        return self.type

    def get_units(self) -> int:
        """
        Returns the number of abstract units in the set.
        @returns number of abstract units
        """
        return self.units

    def is_active(self) -> bool:
        """
        Checks if the resource set is active: allocated units are active.
        @returns true if this ResourceSet is active
        """
        if self.resources is not None:
            return self.resources.is_active()
        return False

    def is_closed(self) -> bool:
        """
        Checks if the resource set is closed: there are no active units. Do not
        call this method unless the set had a close in progress: a set with
        failed units or one that has not yet been activated may register as
        "closed".
        @returns true if this ResourceSet is active
        """
        if self.resources is None:
            return True
        if self.resources.get_units() == 0:
            return True
        return False

    def is_empty(self) -> bool:
        if self.updated is not None and self.updated.get_units() > 0:
            return False
        if self.gained is not None and self.gained.get_units() > 0:
            return False
        if self.lost is not None and self.lost.get_units() > 0:
            return False
        if self.modified is not None and self.modified.get_units() > 0:
            return False
        return True

    def prepare_probe(self):
        """
        Prepares a probe: updates ConcreteSet to reflect underlying resource
        status.
        @raises Exception in case of error
        """
        if self.resources is not None:
            self.resources.probe()
            if self.released is None:
                self.released = self.resources.collect_released()

    def probe(self):
        return

    def service_check(self):
        if self.resources is None:
            raise ResourcesException("Internal Error: WARNING: service post-op call on non-concrete reservation")

    def close(self):
        """
        Initiate close on the concrete resources
        """
        if not self.is_closing:
            self.is_closing = True
            self.resources.close()

    def service_extend(self):
        """
        Complete service for a term extension (server side).
        """
        self.service_check()
        # An elastic reservation can change concrete resources on extend. The
        # modifications are left in update/gained/lost by *Change() above. On
        # agent the concrete is updated synchronously in SoftChange, so this
        # code segment applies to authority only.
        my_gained = self.gained
        self.gained = None
        # An elastic reservation can change concrete resources on extend. The
        # modifications are left in update/gained/lost by *Change() above. On
        # agent the concrete is updated synchronously in SoftChange, so this
        # code segment applies to authority only.
        my_lost = self.lost
        self.lost = None
        # An elastic reservation can change concrete resources on extend. The
        # modifications are left in update/gained/lost by *Change() above. On
        # agent the concrete is updated synchronously in SoftChange, so this
        # code segment applies to authority only.
        my_modified = self.modified
        self.modified = None
        if my_gained is not None:
            self.resources.add(concrete_set=my_gained, configure=True)
        if my_lost is not None:
            self.resources.remove(concrete_set=my_lost, configure=True)
        if my_modified is not None:
            self.resources.modify(concrete_set=my_modified, configure=True)

    def service_modify(self):
        """
        Complete service for a term extension (server side).
        @raises Exception in case of error
        """
        self.service_check()
        self.resources.modify(concrete_set=self.resources, configure=True)

    def service_reserve_site(self):
        cs = None
        if self.gained is not None:
            cs = self.gained
            self.gained = None

        if cs is not None:
            self.resources.add(concrete_set=cs, configure=True)

    def service_update(self, *, reservation: IReservation):
        """
        Service a resource set update. Any changes to existing
        concrete resources should have been left in "updated" by an update
        operation.
        @params reservation: reservation
        @raises Exception in case of error
        """
        cs = None
        if self.updated is not None:
            cs = self.updated
            self.updated = None
        if cs is not None:
            self.resources.change(concrete_set=cs, configure=True)

    def set_config_properties(self, *, p: dict):
        """
        Sets the configuration properties.
        @params p : configuration properties list
        """
        if self.properties is None:
            self.properties = ResourceData()
        self.properties.merge_properties(from_props=p, to_props=self.properties.get_configuration_properties())

    def set_local_properties(self, *, p: dict):
        """
        Sets the local properties.
        @params p : local properties list
        """
        if self.properties is None:
            self.properties = ResourceData()
        self.properties.merge_properties(from_props=p, to_props=self.properties.get_local_properties())

    def set_request_properties(self, *, p: dict):
        """
        Sets the request properties.
        @params p : request properties list
        """
        if self.properties is None:
            self.properties = ResourceData()
        self.properties.merge_properties(from_props=p, to_props=self.properties.get_request_properties())

    def set_reservation_id(self, *, rid: ID):
        """
        Attaches the reservation identifier to the set.
        @params rid reservation identifier
        """
        self.rid = rid

    def set_resource_properties(self, *, p: dict):
        """
        Sets the resource properties.
        @params p : resource properties list
        """
        if self.properties is None:
            self.properties = ResourceData()
        self.properties.merge_properties(from_props=p, to_props=self.properties.get_resource_properties())

    def set_resources(self, *, cset: IConcreteSet):
        """
        Set the concrete resources. Used by proxies.
        @params cset :concrete resource set
        """
        self.resources = cset

    def set_type(self, *, rtype: resource_type):
        """
        Sets the resource type for the set.
        @params rtype : resource type
        """
        self.type = rtype

    def set_units(self, *, units: int):
        """
        Sets the number of abstract units in the set.
        @params units: number of abstract units
        """
        self.units = units

    def setup(self, *, reservation: IReservation):
        """
        Passes information about the containing reservation to the concrete set.
        @params reservation: containing reservation
        """
        if self.resources is not None:
            self.resources.setup(reservation=reservation)

    def __str__(self):
        result = "rset: units=[{}] ".format(self.units)
        if self.resources is not None:
            result += " concrete:[{}]".format(self.resources)
        if self.properties is not None:
            result += " properties: [{}]".format(self.properties)
        return result

    def update(self, *, reservation: IReservation, resource_set: ResourceSet):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.invalid_argument)

        if resource_set.resources is not None:
            self.full_update(reservation=reservation, resource_set=resource_set)
        else:
            self.delta_update(reservation=reservation, resource_set=resource_set)

    def update_properties(self, *, reservation: IReservation, resource_set):
        if reservation is None or resource_set is None:
            raise ResourcesException(Constants.invalid_argument)

        self.merge_properties(reservation=reservation, resource_set=resource_set)

    def validate(self):
        """
        Validates a fresh ResourceSet passed in from outside
        @raises Exception in case of error thrown if the set is determined to be invalid
        """
        if self.units < 0:
            raise ResourcesException("invalid unit count:{}".format(self.units))

    def validate_incoming(self):
        """
        Validates a ResourceSet in an incoming ticket or lease
        request (server) or in an incoming ticket or lease update (client).
        Called for each incoming request/update to check validity with no locks
        held.
        @raises Exception in case of error
        """
        self.validate()
        if self.resources is not None:
            self.resources.validate_incoming()

    def validate_incoming_ticket(self, *, term: Term):
        """
        Validate match between abstract and concrete ResourceSet in a ResourceSet
        representing an incoming ticket.
        @params t : optional term associated with ResourceSet
        @raises Exception in case of error if validation fails
        """
        if self.resources is None:
            if self.units != 0:
                raise ResourcesException("no resources to back incoming ticket")
            return
        if self.resources.get_units() != self.units:
            raise ResourcesException("size mismatch on incoming ticket {} != {}".format(self.resources.get_units(),
                                                                               self.units))
        self.resources.validate_concrete(rtype=self.type, units=self.units, term=term)

    def validate_outgoing(self):
        """
        Validates a ResourceSet that is about to be sent to another
        actor. Client-side only.
        @raises Exception in case of error
        """
        self.validate()
        if self.resources is not None:
            self.resources.validate_outgoing()

    def clone(self):
        rd = self.properties.clone()
        clone = ResourceSet(units=self.units, rtype=self.type, rdata=rd)
        clone.resources = self.resources.clone()
        return clone
